
import codecs
from distutils.version import (
    LooseVersion,
)
import functools
import json

import platone.packages.eth_abi
from platone.packages.eth_abi.abi import (
    process_type,
)
from platone.packages.eth_utils import (
    to_checksum_address,
)
from platone.packages.eth_utils.address import (
    is_binary_address,
)
from hexbytes import (
    HexBytes,
)

from platone.exceptions import (
    InvalidAddress,
)
from platone.utils.encoding import (
    hexstr_if_str,
    text_if_str,
    to_bytes,
    to_hex,
    to_text,
)
from platone.utils.ens import (
    StaticENS,
    is_ens_name,
    validate_name_has_address,
)
from platone.utils.toolz import (
    curry,
)
from platone.utils.validation import (
    validate_abi,
    validate_address,
)


def implicitly_identity(to_wrap):
    @functools.wraps(to_wrap)
    def wrapper(abi_type, data):
        modified = to_wrap(abi_type, data)
        if modified is None:
            return abi_type, data
        else:
            return modified
    return wrapper


#
# Return Normalizers
#


@implicitly_identity
def addresses_checksummed(abi_type, data):
    if abi_type == 'address':
        return abi_type, to_checksum_address(data)


@implicitly_identity
def decode_abi_strings(abi_type, data):
    if abi_type == 'string':
        return abi_type, codecs.decode(data, 'utf8', 'backslashreplace')


#
# Argument Normalizers
#


@implicitly_identity
def abi_bytes_to_hex(abi_type, data):
    base, sub, arrlist = process_type(abi_type)
    if base == 'bytes' and not arrlist:
        bytes_data = hexstr_if_str(to_bytes, data)
        if not sub:
            return abi_type, to_hex(bytes_data)
        else:
            num_bytes = int(sub)
            if len(bytes_data) <= num_bytes:
                padded = bytes_data.ljust(num_bytes, b'\0')
                return abi_type, to_hex(padded)
            else:
                raise ValueError(
                    "This value was expected to be at most %d bytes, but instead was %d: %r" % (
                        (num_bytes, len(bytes_data), data)
                    )
                )


@implicitly_identity
def abi_int_to_hex(abi_type, data):
    base, _sub, arrlist = process_type(abi_type)
    if base == 'uint' and not arrlist:
        return abi_type, hexstr_if_str(to_hex, data)


@implicitly_identity
def abi_string_to_hex(abi_type, data):
    if abi_type == 'string':
        return abi_type, text_if_str(to_hex, data)


@implicitly_identity
def abi_string_to_text(abi_type, data):
    if abi_type == 'string':
        return abi_type, text_if_str(to_text, data)


@implicitly_identity
def abi_bytes_to_bytes(abi_type, data):
    base, sub, arrlist = process_type(abi_type)
    if base == 'bytes' and not arrlist:
        return abi_type, hexstr_if_str(to_bytes, data)


@implicitly_identity
def abi_address_to_hex(abi_type, data):
    if abi_type == 'address':
        validate_address(data)
        if is_binary_address(data):
            return abi_type, to_checksum_address(data)


@curry
def abi_ens_resolver(w3, abi_type, val):
    if abi_type == 'address': #and is_ens_name(val):
        if w3 is None:
            raise InvalidAddress(
                "Could not look up name %r because no web3"
                " connection available" % (val)
            )
        # elif w3.ens is None:
        #     raise InvalidAddress(
        #         "Could not look up name %r because ENS is"
        #         " set to None" % (val)
        #     )
        # elif int(w3.net.version) is not 1 and not isinstance(w3.ens, StaticENS):
        #     raise InvalidAddress(
        #         "Could not look up name %r because web3 is"
        #         " not connected to mainnet" % (val)
        #     )
        else:
            return (abi_type, val)
    else:
        return (abi_type, val)


BASE_RETURN_NORMALIZERS = [
    addresses_checksummed,
]


# if LooseVersion(eth_abi.__version__) < LooseVersion("2"):
BASE_RETURN_NORMALIZERS.append(decode_abi_strings)


#
# Property Normalizers
#


def normalize_abi(abi):
    if isinstance(abi, str):
        abi = json.loads(abi)
    validate_abi(abi)
    return abi


def normalize_address(address):
    # if address:
    #     if is_ens_name(address):
    #         validate_name_has_address(ens, address)
    #     else:
    #         validate_address(address)
    return address
def normalize_vmtype(vmtype):

    return vmtype

def normalize_bytecode(bytecode):
    if bytecode:
        bytecode = HexBytes(bytecode)
    return bytecode



CNS_ABI = [
    {
        "constant": True,
        "inputs": [],
        "name": "GetAllContracts",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "version",
                "type": "string"
            }
        ],
        "name": "GetContractAddress",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "GetContractsByAddr",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "GetContractsByOwner",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "version",
                "type": "string"
            },
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "Register",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "version",
                "type": "string"
            }
        ],
        "name": "Unregister",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

LEDGER_MANAGER_ABI = [
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "ledgerJson",
                "type": "string"
            }
        ],
        "name": "CreateLedger",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "GetAllLedgers",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "ledgerName",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "nodeID",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "blsPubKey",
                "type": "string"
            }
        ],
        "name": "JoinLedger",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "nodeID",
                "type": "string"
            }
        ],
        "name": "JoinedLedgers",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "ledgerName",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "nodeID",
                "type": "string"
            }
        ],
        "name": "QuitLedger",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "ledgerName",
                "type": "string"
            }
        ],
        "name": "TerminateLedger",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

LEDGER_NODE_MANAGER_ABI = [
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "pubKey",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "blsPubKey",
                "type": "string"
            }
        ],
        "name": "Add",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "GrantPerm",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "List",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "pubKey",
                "type": "string"
            }
        ],
        "name": "Remove",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "RevokePerm",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "pubKey",
                "type": "string"
            },
            {
                "internalType": "uint8",
                "name": "nodeType",
                "type": "uint8"
            }
        ],
        "name": "UpdateNodeType",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

NODE_MANAGER_ABI = [
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "publicKey",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "blsPubKey",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "hostAddress",
                "type": "string"
            },
            {
                "internalType": "uint16",
                "name": "rpcPort",
                "type": "uint16"
            },
            {
                "internalType": "uint16",
                "name": "p2pPort",
                "type": "uint16"
            }
        ],
        "name": "Add",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "publicKey",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "blsPubKey",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "hostAddress",
                "type": "string"
            },
            {
                "internalType": "uint16",
                "name": "rpcPort",
                "type": "uint16"
            },
            {
                "internalType": "uint16",
                "name": "p2pPort",
                "type": "uint16"
            }
        ],
        "name": "Apply",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "uint8",
                "name": "auditStat",
                "type": "uint8"
            },
            {
                "internalType": "string",
                "name": "auditReason",
                "type": "string"
            }
        ],
        "name": "Audit",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            }
        ],
        "name": "Delete",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            }
        ],
        "name": "Disable",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            }
        ],
        "name": "Enable",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            }
        ],
        "name": "GetByName",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "publicKey",
                "type": "string"
            }
        ],
        "name": "GetByPublicKey",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "ListAll",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "hostAddress",
                "type": "string"
            }
        ],
        "name": "ListByHostAddress",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            }
        ],
        "name": "ListByOwner",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint8",
                "name": "status",
                "type": "uint8"
            }
        ],
        "name": "ListByStat",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint8",
                "name": "nodeType",
                "type": "uint8"
            }
        ],
        "name": "ListByType",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "hostAddress",
                "type": "string"
            },
            {
                "internalType": "uint16",
                "name": "rpcPort",
                "type": "uint16"
            },
            {
                "internalType": "uint16",
                "name": "p2pPort",
                "type": "uint16"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            }
        ],
        "name": "Update",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "uint8",
                "name": "nodeType",
                "type": "uint8"
            }
        ],
        "name": "UpdateType",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

PARAM_MANAGER_ABI = [
    {
        "constant": False,
        "inputs": [],
        "name": "DisableDeploy",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [],
        "name": "EnableDeploy",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [],
        "name": "GetSystemParameter",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "uint64",
                "name": "blockGasLimit",
                "type": "uint64"
            }
        ],
        "name": "UpdateBlockGasLimit",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "bool",
                "name": "isProduceEmptyBlock",
                "type": "bool"
            }
        ],
        "name": "UpdateIsProduceEmptyBlock",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "bool",
                "name": "isTxUseGas",
                "type": "bool"
            }
        ],
        "name": "UpdateIsTxUseGas",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "uint64",
                "name": "blockGasLimit",
                "type": "uint64"
            },
            {
                "internalType": "uint64",
                "name": "txGasLimit",
                "type": "uint64"
            },
            {
                "internalType": "bool",
                "name": "isTxUseGas",
                "type": "bool"
            },
            {
                "internalType": "bool",
                "name": "isProduceEmptyBlock",
                "type": "bool"
            },
            {
                "internalType": "bool",
                "name": "enableDeploy",
                "type": "bool"
            }
        ],
        "name": "UpdateSystemParameter",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "uint64",
                "name": "txGasLimit",
                "type": "uint64"
            }
        ],
        "name": "UpdateTxGasLimit",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

RE_ENC_ABI = [
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "cipher",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "ownerPk1",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "reEncryptKey",
                "type": "string"
            }
        ],
        "name": "ReEncrypt",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    }
]

USER_MANAGER_ABI = [
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "mobile",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "email",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "Add",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "AddRoles",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "uint8",
                "name": "auditStat",
                "type": "uint8"
            },
            {
                "internalType": "string",
                "name": "auditReason",
                "type": "string"
            }
        ],
        "name": "Audit",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "Delete",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "Disable",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "Enable",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            }
        ],
        "name": "GetByAddr",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            }
        ],
        "name": "GetByName",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "HasRoles",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint32",
                "name": "pageNum",
                "type": "uint32"
            },
            {
                "internalType": "uint32",
                "name": "pageSize",
                "type": "uint32"
            }
        ],
        "name": "PageAll",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            },
            {
                "internalType": "uint32",
                "name": "pageNum",
                "type": "uint32"
            },
            {
                "internalType": "uint32",
                "name": "pageSize",
                "type": "uint32"
            }
        ],
        "name": "PageByRoles",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": True,
        "inputs": [
            {
                "internalType": "uint8",
                "name": "status",
                "type": "uint8"
            },
            {
                "internalType": "uint32",
                "name": "pageNum",
                "type": "uint32"
            },
            {
                "internalType": "uint32",
                "name": "pageSize",
                "type": "uint32"
            }
        ],
        "name": "PageByStat",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "string",
                "name": "name",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "mobile",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "email",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "Register",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "RemoveRoles",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "uint64",
                "name": "roles",
                "type": "uint64"
            }
        ],
        "name": "SetRoles",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "constant": False,
        "inputs": [
            {
                "internalType": "address",
                "name": "addr",
                "type": "address"
            },
            {
                "internalType": "string",
                "name": "mobile",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "email",
                "type": "string"
            },
            {
                "internalType": "string",
                "name": "desc",
                "type": "string"
            }
        ],
        "name": "Update",
        "outputs": [
            {
                "internalType": "int32",
                "name": "",
                "type": "int32"
            }
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function"
    }
]



from platone.module import Module
from platone.utils.builtin_abis import RE_ENC_ABI
from platone.utils.transactions import send_obj_transaction


class ReEnc(Module):
    ABI = RE_ENC_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.re_enc_address, abi=self.ABI,
                                                   ledger=self.web3.system_ledger)

    def reEncrypt(self, cipher, ownerPk1, reEncryptKey, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.ReEncrypt(cipher, ownerPk1, reEncryptKey).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)




from platone.packages.platone_keys.constants import (
    SECPK1_N,
)


def int_to_byte(value: int) -> bytes:
    return bytes([value])


def coerce_low_s(value: int) -> int:
    """Coerce the s component of an ECDSA signature into its low-s form.

    See https://bitcoin.stackexchange.com/questions/83408/in-ecdsa-why-is-r-%E2%88%92s-mod-n-complementary-to-r-s  # noqa: W501
    or https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2.md.
    """
    return min(value, -value % SECPK1_N)



from collections import (
    namedtuple,
)
import itertools
import re

from platone.packages.eth_abi import (
    is_encodable as eth_abi_is_encodable,
)
from platone.packages.eth_abi.abi import (
    collapse_type,
    process_type,
)
from platone.packages.eth_utils import (
    is_hex,
    is_list_like,
    to_bytes,
    to_text,
    to_tuple,
)

from platone.exceptions import (
    FallbackNotFound,
)
from platone.utils.ens import (
    is_ens_name,
)
from platone.utils.formatters import (
    recursive_map,
)
from platone.utils.toolz import (
    curry,
    partial,
    pipe,
)


def filter_by_type(_type, contract_abi):
    return [abi for abi in contract_abi if abi['type'] == _type]


def filter_by_name(name, contract_abi):
    return [
        abi
        for abi
        in contract_abi
        if (
            abi['type'] not in ('fallback', 'constructor','receive') and
            abi['name'] == name
        )
    ]


def get_abi_input_types(abi):
    if 'inputs' not in abi and abi['type'] == 'fallback':
        return []
    else:
        return [arg['type'] for arg in abi['inputs']]


def get_abi_output_types(abi,vmtype):
    if abi['type'] == 'fallback':
        return []
    else:
        if isinstance(abi['outputs'],dict):
            if vmtype==1:
               return abi['outputs']
            else :
                return abi['outputs']['type']
        else :
            return [arg['type'] for arg in abi['outputs']]


def get_abi_input_names(abi):
    if 'inputs' not in abi and abi['type'] == 'fallback':
        return []
    else:
        return [arg['name'] for arg in abi['inputs']]


def get_fallback_func_abi(contract_abi):
    fallback_abis = filter_by_type('fallback', contract_abi)
    if fallback_abis:
        return fallback_abis[0]
    else:
        raise FallbackNotFound("No fallback function was found in the contract ABI.")


def fallback_func_abi_exists(contract_abi):
    return filter_by_type('fallback', contract_abi)


def get_indexed_event_inputs(event_abi):
    return [arg for arg in event_abi['inputs'] if arg['indexed'] is True]


def exclude_indexed_event_inputs(event_abi):
    return [arg for arg in event_abi['inputs'] if arg['indexed'] is False]


def filter_by_argument_count(num_arguments, contract_abi):
    return [
        abi
        for abi
        in contract_abi
        if len(abi['inputs']) == num_arguments
    ]


def filter_by_argument_name(argument_names, contract_abi):
    return [
        abi
        for abi in contract_abi
        if set(argument_names).intersection(
            get_abi_input_names(abi)
        ) == set(argument_names)
    ]


def is_encodable(_type, value):
    try:
        base, sub, arrlist = _type
    except ValueError:
        base, sub, arrlist = process_type(_type)

    if arrlist:
        if not is_list_like(value):
            return False
        if arrlist[-1] and len(value) != arrlist[-1][0]:
            return False
        sub_type = (base, sub, arrlist[:-1])
        return all(is_encodable(sub_type, sub_value) for sub_value in value)
    elif base == 'address' and is_ens_name(value):
        # ENS names can be used anywhere an address is needed
        # Web3.py will resolve the name to an address before encoding it
        return True
    elif base == 'bytes' and isinstance(value, str):
        # Hex-encoded bytes values can be used anywhere a bytes value is needed
        if is_hex(value) and len(value) % 2 == 0:
            # Require hex-encoding of full bytes (even length)
            bytes_val = to_bytes(hexstr=value)
            return eth_abi_is_encodable(_type, bytes_val)
        else:
            return False
    elif base == 'string' and isinstance(value, bytes):
        # bytes that were encoded with utf-8 can be used anywhere a string is needed
        try:
            string_val = to_text(value)
        except UnicodeDecodeError:
            return False
        else:
            return eth_abi_is_encodable(_type, string_val)
    else:
        return eth_abi_is_encodable(_type, value)


def filter_by_encodability(args, kwargs, contract_abi):
    return [
        function_abi
        for function_abi
        in contract_abi
        if check_if_arguments_can_be_encoded(function_abi, args, kwargs)
    ]


def check_if_arguments_can_be_encoded(function_abi, args, kwargs):
    try:
        arguments = merge_args_and_kwargs(function_abi, args, kwargs)
    except TypeError:
        return False

    if len(function_abi.get('inputs', [])) != len(arguments):
        return False

    types = get_abi_input_types(function_abi)

    return all(
        is_encodable(_type, arg)
        for _type, arg in zip(types, arguments)
    )


def merge_args_and_kwargs(function_abi, args, kwargs):
    if len(args) + len(kwargs) != len(function_abi.get('inputs', [])):
        raise TypeError(
            "Incorrect argument count.  Expected '{0}'.  Got '{1}'".format(
                len(function_abi['inputs']),
                len(args) + len(kwargs),
            )
        )

    if not kwargs:
        return args

    args_as_kwargs = {
        arg_abi['name']: arg
        for arg_abi, arg in zip(function_abi['inputs'], args)
    }
    duplicate_keys = set(args_as_kwargs).intersection(kwargs.keys())
    if duplicate_keys:
        raise TypeError(
            "{fn_name}() got multiple values for argument(s) '{dups}'".format(
                fn_name=function_abi['name'],
                dups=', '.join(duplicate_keys),
            )
        )

    sorted_arg_names = [arg_abi['name'] for arg_abi in function_abi['inputs']]

    unknown_kwargs = {key for key in kwargs.keys() if key not in sorted_arg_names}
    if unknown_kwargs:
        if function_abi.get('name'):
            raise TypeError(
                "{fn_name}() got unexpected keyword argument(s) '{dups}'".format(
                    fn_name=function_abi.get('name'),
                    dups=', '.join(unknown_kwargs),
                )
            )
        # show type instead of name in the error message incase key 'name' is missing.
        raise TypeError(
            "Type: '{_type}' got unexpected keyword argument(s) '{dups}'".format(
                _type=function_abi.get('type'),
                dups=', '.join(unknown_kwargs),
            )
        )

    sorted_args = list(zip(
        *sorted(
            itertools.chain(kwargs.items(), args_as_kwargs.items()),
            key=lambda kv: sorted_arg_names.index(kv[0])
        )
    ))
    if sorted_args:
        return sorted_args[1]
    else:
        return tuple()


def get_constructor_abi(contract_abi):
    candidates = [
        abi for abi in contract_abi if abi['type'] == 'constructor'
    ]
    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        return None
    elif len(candidates) > 1:
        raise ValueError("Found multiple constructors.")


DYNAMIC_TYPES = ['bytes', 'string']

INT_SIZES = range(8, 257, 8)
BYTES_SIZES = range(1, 33)
UINT_TYPES = ['uint{0}'.format(i) for i in INT_SIZES]
INT_TYPES = ['int{0}'.format(i) for i in INT_SIZES]
BYTES_TYPES = ['bytes{0}'.format(i) for i in BYTES_SIZES] + ['bytes32.byte']

STATIC_TYPES = list(itertools.chain(
    ['address', 'bool'],
    UINT_TYPES,
    INT_TYPES,
    BYTES_TYPES,
))

BASE_TYPE_REGEX = '|'.join((
    _type + '(?![a-z0-9])'
    for _type
    in itertools.chain(STATIC_TYPES, DYNAMIC_TYPES)
))

SUB_TYPE_REGEX = (
    r'\['
    '[0-9]*'
    r'\]'
)

TYPE_REGEX = (
    '^'
    '(?:{base_type})'
    '(?:(?:{sub_type})*)?'
    '$'
).format(
    base_type=BASE_TYPE_REGEX,
    sub_type=SUB_TYPE_REGEX,
)


def is_recognized_type(abi_type):
    return bool(re.match(TYPE_REGEX, abi_type))


def is_bool_type(abi_type):
    return abi_type == 'bool'


def is_uint_type(abi_type):
    return abi_type in UINT_TYPES


def is_int_type(abi_type):
    return abi_type in INT_TYPES


def is_address_type(abi_type):
    return abi_type == 'address'


def is_bytes_type(abi_type):
    return abi_type in BYTES_TYPES + ['bytes']


def is_string_type(abi_type):
    return abi_type == 'string'


def size_of_type(abi_type):
    """
    Returns size in bits of abi_type
    """
    if 'string' in abi_type:
        return None
    if 'byte' in abi_type:
        return None
    if '[' in abi_type:
        return None
    if abi_type == 'bool':
        return 8
    if abi_type == 'address':
        return 160
    return int(re.sub(r"\D", "", abi_type))


END_BRACKETS_OF_ARRAY_TYPE_REGEX = r"\[[^]]*\]$"


def sub_type_of_array_type(abi_type):
    if not is_array_type(abi_type):
        raise ValueError(
            "Cannot parse subtype of nonarray abi-type: {0}".format(abi_type)
        )

    return re.sub(END_BRACKETS_OF_ARRAY_TYPE_REGEX, '', abi_type, 1)


def length_of_array_type(abi_type):
    if not is_array_type(abi_type):
        raise ValueError(
            "Cannot parse length of nonarray abi-type: {0}".format(abi_type)
        )

    inner_brackets = re.search(END_BRACKETS_OF_ARRAY_TYPE_REGEX, abi_type).group(0).strip("[]")
    if not inner_brackets:
        return None
    else:
        return int(inner_brackets)


ARRAY_REGEX = (
    "^"
    "[a-zA-Z0-9_]+"
    "({sub_type})+"
    "$"
).format(sub_type=SUB_TYPE_REGEX)


def is_array_type(abi_type):
    return bool(re.match(ARRAY_REGEX, abi_type))


NAME_REGEX = (
    '[a-zA-Z_]'
    '[a-zA-Z0-9_]*'
)


ENUM_REGEX = (
    '^'
    '{lib_name}'
    r'\.'
    '{enum_name}'
    '$'
).format(lib_name=NAME_REGEX, enum_name=NAME_REGEX)


def is_probably_enum(abi_type):
    return bool(re.match(ENUM_REGEX, abi_type))


@to_tuple
def normalize_event_input_types(abi_args):
    for arg in abi_args:
        if is_recognized_type(arg['type']):
            yield arg
        elif is_probably_enum(arg['type']):
            yield {k: 'uint8' if k == 'type' else v for k, v in arg.items()}
        else:
            yield arg


def abi_to_signature(abi):
    function_signature = "{fn_name}({fn_input_types})".format(
        fn_name=abi['name'],
        fn_input_types=','.join([
            arg['type'] for arg in normalize_event_input_types(abi.get('inputs', []))
        ]),
    )
    return function_signature


########################################################
#
#  Conditionally modifying data, tagged with ABI Types
#
########################################################


@curry
def map_abi_data(normalizers, types, data):
    '''
    This function will apply normalizers to your data, in the
    context of the relevant types. Each normalizer is in the format:

    def normalizer(datatype, data):
        # Conditionally modify data
        return (datatype, data)

    Where datatype is a valid ABI type string, like "uint".

    In case of an array, like "bool[2]", normalizer will receive `data`
    as an iterable of typed data, like `[("bool", True), ("bool", False)]`.

    Internals
    ---

    This is accomplished by:

    1. Decorating the data tree with types
    2. Recursively mapping each of the normalizers to the data
    3. Stripping the types back out of the tree
    '''
    pipeline = itertools.chain(
        [abi_data_tree(types)],
        map(data_tree_map, normalizers),
        [partial(recursive_map, strip_abi_type)],
    )

    return pipe(data, *pipeline)


@curry
def abi_data_tree(types, data):
    '''
    Decorate the data tree with pairs of (type, data). The pair tuple is actually an
    ABITypedData, but can be accessed as a tuple.

    As an example:

    >>> abi_data_tree(types=["bool[2]", "uint"], data=[[True, False], 0])
    [("bool[2]", [("bool", True), ("bool", False)]), ("uint256", 0)]
    '''
    return [
        abi_sub_tree(data_type, data_value)
        for data_type, data_value
        in zip(types, data)
    ]


@curry
def data_tree_map(func, data_tree):
    '''
    Map func to every ABITypedData element in the tree. func will
    receive two args: abi_type, and data
    '''
    def map_to_typed_data(elements):
        if isinstance(elements, ABITypedData) and elements.abi_type is not None:
            return ABITypedData(func(*elements))
        else:
            return elements
    return recursive_map(map_to_typed_data, data_tree)


class ABITypedData(namedtuple('ABITypedData', 'abi_type, data')):
    '''
    This class marks data as having a certain ABI-type.

    >>> a1 = ABITypedData(['address', addr1])
    >>> a2 = ABITypedData(['address', addr2])
    >>> addrs = ABITypedData(['address[]', [a1, a2])

    You can access the fields using tuple() interface, or with
    attributes:

    >>> assert a1.abi_type == a1[0]
    >>> assert a1.data == a1[1]

    Unlike a typical `namedtuple`, you initialize with a single
    positional argument that is iterable, to match the init
    interface of all other relevant collections.
    '''
    def __new__(cls, iterable):
        return super().__new__(cls, *iterable)


def abi_sub_tree(data_type, data_value):
    if data_type is None:
        return ABITypedData([None, data_value])

    try:
        base, sub, arrlist = data_type
    except ValueError:
        base, sub, arrlist = process_type(data_type)

    collapsed = collapse_type(base, sub, arrlist)

    if arrlist:
        sub_type = (base, sub, arrlist[:-1])
        return ABITypedData([
            collapsed,
            [
                abi_sub_tree(sub_type, sub_value)
                for sub_value in data_value
            ],
        ])
    else:
        return ABITypedData([collapsed, data_value])


def strip_abi_type(elements):
    if isinstance(elements, ABITypedData):
        return elements.data
    else:
        return elements



import pytest

from platone.packages.eth_utils import (
    is_boolean,
    is_integer,
    is_string,
)


class NetModuleTest:
    def test_net_version(self, web3):
        version = web3.net.version

        assert is_string(version)
        assert version.isdigit()

    def test_net_listening(self, web3):
        listening = web3.net.listening

        assert is_boolean(listening)

    def test_net_peerCount(self, web3):
        peer_count = web3.net.peerCount

        assert is_integer(peer_count)

    def test_net_chainId(self, web3):
        with pytest.warns(DeprecationWarning):
            web3.net.chainId



#-*-coding:utf-8-*-
import copy
from .func import xor, rotl, get_uint32_be, put_uint32_be, \
        bytes_to_list, list_to_bytes, padding, unpadding

#Expanded SM4 box table
SM4_BOXES_TABLE = [
    0xd6,0x90,0xe9,0xfe,0xcc,0xe1,0x3d,0xb7,0x16,0xb6,0x14,0xc2,0x28,0xfb,0x2c,
    0x05,0x2b,0x67,0x9a,0x76,0x2a,0xbe,0x04,0xc3,0xaa,0x44,0x13,0x26,0x49,0x86,
    0x06,0x99,0x9c,0x42,0x50,0xf4,0x91,0xef,0x98,0x7a,0x33,0x54,0x0b,0x43,0xed,
    0xcf,0xac,0x62,0xe4,0xb3,0x1c,0xa9,0xc9,0x08,0xe8,0x95,0x80,0xdf,0x94,0xfa,
    0x75,0x8f,0x3f,0xa6,0x47,0x07,0xa7,0xfc,0xf3,0x73,0x17,0xba,0x83,0x59,0x3c,
    0x19,0xe6,0x85,0x4f,0xa8,0x68,0x6b,0x81,0xb2,0x71,0x64,0xda,0x8b,0xf8,0xeb,
    0x0f,0x4b,0x70,0x56,0x9d,0x35,0x1e,0x24,0x0e,0x5e,0x63,0x58,0xd1,0xa2,0x25,
    0x22,0x7c,0x3b,0x01,0x21,0x78,0x87,0xd4,0x00,0x46,0x57,0x9f,0xd3,0x27,0x52,
    0x4c,0x36,0x02,0xe7,0xa0,0xc4,0xc8,0x9e,0xea,0xbf,0x8a,0xd2,0x40,0xc7,0x38,
    0xb5,0xa3,0xf7,0xf2,0xce,0xf9,0x61,0x15,0xa1,0xe0,0xae,0x5d,0xa4,0x9b,0x34,
    0x1a,0x55,0xad,0x93,0x32,0x30,0xf5,0x8c,0xb1,0xe3,0x1d,0xf6,0xe2,0x2e,0x82,
    0x66,0xca,0x60,0xc0,0x29,0x23,0xab,0x0d,0x53,0x4e,0x6f,0xd5,0xdb,0x37,0x45,
    0xde,0xfd,0x8e,0x2f,0x03,0xff,0x6a,0x72,0x6d,0x6c,0x5b,0x51,0x8d,0x1b,0xaf,
    0x92,0xbb,0xdd,0xbc,0x7f,0x11,0xd9,0x5c,0x41,0x1f,0x10,0x5a,0xd8,0x0a,0xc1,
    0x31,0x88,0xa5,0xcd,0x7b,0xbd,0x2d,0x74,0xd0,0x12,0xb8,0xe5,0xb4,0xb0,0x89,
    0x69,0x97,0x4a,0x0c,0x96,0x77,0x7e,0x65,0xb9,0xf1,0x09,0xc5,0x6e,0xc6,0x84,
    0x18,0xf0,0x7d,0xec,0x3a,0xdc,0x4d,0x20,0x79,0xee,0x5f,0x3e,0xd7,0xcb,0x39,
    0x48,
]

# System parameter
SM4_FK = [0xa3b1bac6,0x56aa3350,0x677d9197,0xb27022dc]

# fixed parameter
SM4_CK = [
    0x00070e15,0x1c232a31,0x383f464d,0x545b6269,
    0x70777e85,0x8c939aa1,0xa8afb6bd,0xc4cbd2d9,
    0xe0e7eef5,0xfc030a11,0x181f262d,0x343b4249,
    0x50575e65,0x6c737a81,0x888f969d,0xa4abb2b9,
    0xc0c7ced5,0xdce3eaf1,0xf8ff060d,0x141b2229,
    0x30373e45,0x4c535a61,0x686f767d,0x848b9299,
    0xa0a7aeb5,0xbcc3cad1,0xd8dfe6ed,0xf4fb0209,
    0x10171e25,0x2c333a41,0x484f565d,0x646b7279
]

SM4_ENCRYPT = 0
SM4_DECRYPT = 1


class CryptSM4(object):

    def __init__(self, mode=SM4_ENCRYPT):
        self.sk = [0]*32
        self.mode = mode
    # Calculating round encryption key.
    # args:    [in] a: a is a 32 bits unsigned value;
    # return: sk[i]: i{0,1,2,3,...31}.
    @classmethod
    def _round_key(cls, ka):
        b = [0, 0, 0, 0]
        a = put_uint32_be(ka)
        b[0] = SM4_BOXES_TABLE[a[0]]
        b[1] = SM4_BOXES_TABLE[a[1]]
        b[2] = SM4_BOXES_TABLE[a[2]]
        b[3] = SM4_BOXES_TABLE[a[3]]
        bb = get_uint32_be(b[0:4])
        rk = bb ^ (rotl(bb, 13)) ^ (rotl(bb, 23))
        return rk

    # Calculating and getting encryption/decryption contents.
    # args:    [in] x0: original contents;
    # args:    [in] x1: original contents;
    # args:    [in] x2: original contents;
    # args:    [in] x3: original contents;
    # args:    [in] rk: encryption/decryption key;
    # return the contents of encryption/decryption contents.
    @classmethod
    def _f(cls, x0, x1, x2, x3, rk):
        # "T algorithm" == "L algorithm" + "t algorithm".
        # args:    [in] a: a is a 32 bits unsigned value;
        # return: c: c is calculated with line algorithm "L" and nonline algorithm "t"
        def _sm4_l_t(ka):
            b = [0, 0, 0, 0]
            a = put_uint32_be(ka)
            b[0] = SM4_BOXES_TABLE[a[0]]
            b[1] = SM4_BOXES_TABLE[a[1]]
            b[2] = SM4_BOXES_TABLE[a[2]]
            b[3] = SM4_BOXES_TABLE[a[3]]
            bb = get_uint32_be(b[0:4])
            c = bb ^ (rotl(bb, 2)) ^ (rotl(bb, 10)) ^ (rotl(bb, 18)) ^ (rotl(bb, 24))
            return c
        return (x0 ^ _sm4_l_t(x1 ^ x2 ^ x3 ^ rk))

    def set_key(self, key, mode):
        key = bytes_to_list(key)
        MK = [0, 0, 0, 0]
        k = [0]*36
        MK[0] = get_uint32_be(key[0:4])
        MK[1] = get_uint32_be(key[4:8])
        MK[2] = get_uint32_be(key[8:12])
        MK[3] = get_uint32_be(key[12:16])
        k[0:4] = xor(MK[0:4], SM4_FK[0:4])
        for i in range(32):
            k[i + 4] = k[i] ^ (
                self._round_key(k[i + 1] ^ k[i + 2] ^ k[i + 3] ^ SM4_CK[i]))
            self.sk[i] = k[i + 4]
        self.mode = mode
        if mode == SM4_DECRYPT:
            for idx in range(16):
                t = self.sk[idx]
                self.sk[idx] = self.sk[31 - idx]
                self.sk[31 - idx] = t

    def one_round(self, sk, in_put):
        out_put = []
        ulbuf = [0]*36
        ulbuf[0] = get_uint32_be(in_put[0:4])
        ulbuf[1] = get_uint32_be(in_put[4:8])
        ulbuf[2] = get_uint32_be(in_put[8:12])
        ulbuf[3] = get_uint32_be(in_put[12:16])
        for idx in range(32):
            ulbuf[idx + 4] = self._f(ulbuf[idx], ulbuf[idx + 1], ulbuf[idx + 2], ulbuf[idx + 3], sk[idx])

        out_put += put_uint32_be(ulbuf[35])
        out_put += put_uint32_be(ulbuf[34])
        out_put += put_uint32_be(ulbuf[33])
        out_put += put_uint32_be(ulbuf[32])
        return out_put

    def crypt_ecb(self, input_data):
        # SM4-ECB block encryption/decryption
        input_data = bytes_to_list(input_data)
        if self.mode == SM4_ENCRYPT:
            input_data = padding(input_data)
        length = len(input_data)
        i = 0
        output_data = []
        while length > 0:
            output_data += self.one_round(self.sk, input_data[i:i+16])
            i += 16
            length -= 16
        if self.mode == SM4_DECRYPT:
            return list_to_bytes(unpadding(output_data))
        return list_to_bytes(output_data)

    def crypt_cbc(self, iv, input_data):
        #SM4-CBC buffer encryption/decryption
        i = 0
        output_data = []
        tmp_input = [0]*16
        iv = bytes_to_list(iv)
        if self.mode == SM4_ENCRYPT:
            input_data = padding(bytes_to_list(input_data))
            length = len(input_data)
            while length > 0:
                tmp_input[0:16] = xor(input_data[i:i+16], iv[0:16])
                output_data += self.one_round(self.sk, tmp_input[0:16])
                iv = copy.deepcopy(output_data[i:i+16])
                i += 16
                length -= 16
            return list_to_bytes(output_data)
        else:
            length = len(input_data)
            while length > 0:
                output_data += self.one_round(self.sk, input_data[i:i+16])
                output_data[i:i+16] = xor(output_data[i:i+16], iv[0:16])
                iv = copy.deepcopy(input_data[i:i + 16])
                i += 16
                length -= 16
            return list_to_bytes(unpadding(output_data))



from itertools import accumulate
import abc
import codecs
import decimal

from platone.packages.eth_utils import (
    int_to_big_endian,
    is_boolean,
    is_integer,
    is_number,
    is_address,
    is_bytes,
    is_text,
    is_list_like,
    to_canonical_address,
)

from platone.packages.eth_abi.base import (
    BaseCoder,
    parse_type_str,
    parse_tuple_type_str,
)

from platone.packages.eth_abi.exceptions import (
    EncodingTypeError,
    IllegalValue,
    ValueOutOfBounds,
)

from platone.packages.eth_abi.utils.numeric import (
    TEN,
    abi_decimal_context,
    compute_signed_integer_bounds,
    compute_unsigned_integer_bounds,
    compute_signed_fixed_bounds,
    compute_unsigned_fixed_bounds,
    compute_signed_real_bounds,
    compute_unsigned_real_bounds,
    ceil32,
)
from platone.packages.eth_abi.utils.padding import (
    fpad,
    zpad,
    zpad_right,
)


class BaseEncoder(BaseCoder, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encode(self, value):  # pragma: no cover
        """
        Encodes the given Python value as a sequence of bytes.
        """
        pass

    @abc.abstractmethod
    def validate_value(self, value):
        """
        Determines whether or not the given value can be encoded by this
        encoder.
        """
        pass

    def __call__(self, value):
        return self.encode(value)


class TupleEncoder(BaseEncoder):
    encoders = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.is_dynamic = any(e.is_dynamic for e in self.encoders)

    def validate(self):
        super().validate()

        if self.encoders is None:
            raise ValueError("`encoders` may not be none")

    def validate_value(self, value):
        if not is_list_like(value):
            raise EncodingTypeError(
                "Cannot encode value of type {0} using tuple encoder.  Must be "
                "a list-like object such as an array or tuple".format(
                    type(value),
                )
            )

        if len(value) != len(self.encoders):
            raise ValueOutOfBounds(
                "Expected value with length {0}.  Provided value has {1} "
                "elements".format(len(self.encoders), len(value))
            )

        for item, encoder in zip(value, self.encoders):
            encoder.validate_value(item)

    def encode(self, values):
        self.validate_value(values)

        raw_head_chunks = []
        tail_chunks = []
        for value, encoder in zip(values, self.encoders):
            if encoder.is_dynamic:
                raw_head_chunks.append(None)
                tail_chunks.append(encoder(value))
            else:
                raw_head_chunks.append(encoder(value))
                tail_chunks.append(b'')

        head_length = sum(
            32 if item is None else len(item)
            for item in raw_head_chunks
        )
        tail_offsets = (0,) + tuple(accumulate(map(len, tail_chunks[:-1])))
        head_chunks = tuple(
            encode_uint_256(head_length + offset) if chunk is None else chunk
            for chunk, offset in zip(raw_head_chunks, tail_offsets)
        )

        encoded_value = b''.join(head_chunks + tuple(tail_chunks))

        return encoded_value

    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry):
        encoders = tuple(registry.get_encoder(str(c)) for c in abi_type.components)

        return cls(encoders=encoders)


class FixedSizeEncoder(BaseEncoder):
    value_bit_size = None
    data_byte_size = None
    encode_fn = None
    type_check_fn = None
    is_big_endian = None

    def validate(self):
        super().validate()

        if self.value_bit_size is None:
            raise ValueError("`value_bit_size` may not be none")
        if self.data_byte_size is None:
            raise ValueError("`data_byte_size` may not be none")
        if self.encode_fn is None:
            raise ValueError("`encode_fn` may not be none")
        if self.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be none")

        if self.value_bit_size % 8 != 0:
            raise ValueError(
                "Invalid value bit size: {0}.  Must be a multiple of 8".format(
                    self.value_bit_size,
                )
            )

        if self.value_bit_size > self.data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    def validate_value(self, value):
        raise NotImplementedError("Must be implemented by subclasses")

    def encode(self, value):
        self.validate_value(value)
        base_encoded_value = self.encode_fn(value)

        if self.is_big_endian:
            padded_encoded_value = zpad(base_encoded_value, self.data_byte_size)
        else:
            padded_encoded_value = zpad_right(base_encoded_value, self.data_byte_size)

        return padded_encoded_value


class Fixed32ByteSizeEncoder(FixedSizeEncoder):
    data_byte_size = 32


class BooleanEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 8
    is_big_endian = True

    @classmethod
    def validate_value(cls, value):
        if not is_boolean(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {1}".format(
                    type(value),
                    cls.__name__,
                )
            )

    @classmethod
    def encode_fn(cls, value):
        if value is True:
            return b'\x01'
        elif value is False:
            return b'\x00'
        else:
            raise ValueError("Invariant")

    @parse_type_str('bool')
    def from_type_str(cls, abi_type, registry):
        return cls()


class NumberEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = True
    bounds_fn = None
    illegal_value_fn = None
    type_check_fn = None

    def validate(self):
        super().validate()

        if self.bounds_fn is None:
            raise ValueError("`bounds_fn` cannot be null")
        if self.type_check_fn is None:
            raise ValueError("`type_check_fn` cannot be null")

    def validate_value(self, value):
        cls = type(self)
        if not self.type_check_fn(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {1}".format(
                    type(value),
                    cls.__name__,
                )
            )

        illegal_value = (
            self.illegal_value_fn is not None
            and self.illegal_value_fn(value)
        )
        if illegal_value:
            raise IllegalValue(
                'Value {} cannot be encoded by {}'.format(repr(value), cls.__name__)
            )

        lower_bound, upper_bound = self.bounds_fn(self.value_bit_size)
        if value < lower_bound or value > upper_bound:
            raise ValueOutOfBounds(
                "Value {0} cannot be encoded in {1} bits.  Must be bounded "
                "between [{2}, {3}]".format(
                    repr(value),
                    self.value_bit_size,
                    lower_bound,
                    upper_bound,
                )
            )


class UnsignedIntegerEncoder(NumberEncoder):
    encode_fn = staticmethod(int_to_big_endian)
    bounds_fn = staticmethod(compute_unsigned_integer_bounds)
    type_check_fn = staticmethod(is_integer)

    @parse_type_str('uint')
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


encode_uint_256 = UnsignedIntegerEncoder(value_bit_size=256, data_byte_size=32)


class SignedIntegerEncoder(NumberEncoder):
    bounds_fn = staticmethod(compute_signed_integer_bounds)
    type_check_fn = staticmethod(is_integer)

    def encode_fn(self, value):
        return int_to_big_endian(value % (2 ** self.value_bit_size))

    def encode(self, value):
        self.validate_value(value)
        base_encoded_value = self.encode_fn(value)

        if value >= 0:
            padded_encoded_value = zpad(base_encoded_value, self.data_byte_size)
        else:
            padded_encoded_value = fpad(base_encoded_value, self.data_byte_size)

        return padded_encoded_value

    @parse_type_str('int')
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


class BaseFixedEncoder(NumberEncoder):
    frac_places = None

    @staticmethod
    def type_check_fn(value):
        return is_number(value) and not isinstance(value, float)

    @staticmethod
    def illegal_value_fn(value):
        if isinstance(value, decimal.Decimal):
            return value.is_nan() or value.is_infinite()

        return False

    def validate_value(self, value):
        super().validate_value(value)

        with decimal.localcontext(abi_decimal_context):
            residue = value % (TEN ** -self.frac_places)

        if residue > 0:
            raise IllegalValue(
                '{} cannot encode value {}: '
                'residue {} outside allowed fractional precision of {}'.format(
                    type(self).__name__,
                    repr(value),
                    repr(residue),
                    self.frac_places,
                )
            )

    def validate(self):
        super().validate()

        if self.frac_places is None:
            raise ValueError("must specify `frac_places`")

        if self.frac_places <= 0 or self.frac_places > 80:
            raise ValueError("`frac_places` must be in range (0, 80]")


class UnsignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_unsigned_fixed_bounds(self.value_bit_size, self.frac_places)

    def encode_fn(self, value):
        with decimal.localcontext(abi_decimal_context):
            scaled_value = value * TEN ** self.frac_places
            integer_value = int(scaled_value)

        return int_to_big_endian(integer_value)

    @parse_type_str('ufixed')
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            frac_places=frac_places,
        )


class SignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_signed_fixed_bounds(self.value_bit_size, self.frac_places)

    def encode_fn(self, value):
        with decimal.localcontext(abi_decimal_context):
            scaled_value = value * TEN ** self.frac_places
            integer_value = int(scaled_value)

        unsigned_integer_value = integer_value % (2 ** self.value_bit_size)

        return int_to_big_endian(unsigned_integer_value)

    def encode(self, value):
        self.validate_value(value)
        base_encoded_value = self.encode_fn(value)

        if value >= 0:
            padded_encoded_value = zpad(base_encoded_value, self.data_byte_size)
        else:
            padded_encoded_value = fpad(base_encoded_value, self.data_byte_size)

        return padded_encoded_value

    @parse_type_str('fixed')
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            frac_places=frac_places,
        )


class BaseRealEncoder(NumberEncoder):
    low_bit_size = None
    high_bit_size = None
    type_check_fn = staticmethod(is_number)

    def validate(self):
        super().validate()

        if self.high_bit_size is None:
            raise ValueError("`high_bit_size` cannot be null")
        if self.low_bit_size is None:
            raise ValueError("`low_bit_size` cannot be null")
        if self.low_bit_size + self.high_bit_size != self.value_bit_size:
            raise ValueError("high and low bitsizes must sum to the value_bit_size")


class UnsignedRealEncoder(BaseRealEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_unsigned_real_bounds(self.high_bit_size, self.low_bit_size)

    def encode_fn(self, value):
        scaled_value = value * 2 ** self.low_bit_size
        integer_value = int(scaled_value)
        return int_to_big_endian(integer_value)

    @parse_type_str('ureal')
    def from_type_str(cls, abi_type, registry):
        high_bit_size, low_bit_size = abi_type.sub

        return cls(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )


class SignedRealEncoder(BaseRealEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_signed_real_bounds(self.high_bit_size, self.low_bit_size)

    def encode_fn(self, value):
        scaled_value = value * 2 ** self.low_bit_size
        integer_value = int(scaled_value)
        unsigned_integer_value = integer_value % (2 ** (self.high_bit_size + self.low_bit_size))

        return int_to_big_endian(unsigned_integer_value)

    def encode(self, value):
        self.validate_value(value)
        base_encoded_value = self.encode_fn(value)

        if value >= 0:
            padded_encoded_value = zpad(base_encoded_value, self.data_byte_size)
        else:
            padded_encoded_value = fpad(base_encoded_value, self.data_byte_size)

        return padded_encoded_value

    @parse_type_str('real')
    def from_type_str(cls, abi_type, registry):
        high_bit_size, low_bit_size = abi_type.sub

        return cls(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )


class AddressEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 20 * 8
    encode_fn = staticmethod(to_canonical_address)
    is_big_endian = True

    @classmethod
    def validate_value(cls, value):
        if not is_address(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {1}".format(
                    type(value),
                    cls.__name__,
                )
            )

    def validate(self):
        super().validate()

        if self.value_bit_size != 20 * 8:
            raise ValueError('Addresses must be 160 bits in length')

    @parse_type_str('address')
    def from_type_str(cls, abi_type, registry):
        return cls()


class BytesEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = False

    def validate_value(self, value):
        if not is_bytes(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {1}".format(
                    type(value),
                    type(self).__name__,
                )
            )
        if len(value) > self.value_bit_size // 8:
            raise ValueOutOfBounds(
                "String {0} exceeds total byte size for bytes{1} encoding".format(
                    value,
                    self.value_bit_size // 8,
                )
            )

    @staticmethod
    def encode_fn(value):
        return value

    @parse_type_str('bytes')
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub * 8)


class ByteStringEncoder(BaseEncoder):
    is_dynamic = True

    @classmethod
    def validate_value(cls, value):
        if not is_bytes(value):
            raise EncodingTypeError(
                "Value of type {} cannot be encoded by {}".format(
                    type(value),
                    cls.__name__,
                )
            )

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)

        if not value:
            padded_value = b'\x00' * 32
        else:
            padded_value = zpad_right(value, ceil32(len(value)))

        encoded_size = encode_uint_256(len(value))
        encoded_value = encoded_size + padded_value

        return encoded_value

    @parse_type_str('bytes')
    def from_type_str(cls, abi_type, registry):
        return cls()


class TextStringEncoder(BaseEncoder):
    is_dynamic = True

    @classmethod
    def validate_value(cls, value):
        if not is_text(value):
            raise EncodingTypeError(
                "Value of type {} cannot be encoded by {}".format(
                    type(value),
                    cls.__name__,
                )
            )

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)

        value_as_bytes = codecs.encode(value, 'utf8')

        if not value_as_bytes:
            padded_value = b'\x00' * 32
        else:
            padded_value = zpad_right(value_as_bytes, ceil32(len(value_as_bytes)))

        encoded_size = encode_uint_256(len(value_as_bytes))
        encoded_value = encoded_size + padded_value

        return encoded_value

    @parse_type_str('string')
    def from_type_str(cls, abi_type, registry):
        return cls()


class BaseArrayEncoder(BaseEncoder):
    item_encoder = None

    def validate(self):
        super().validate()

        if self.item_encoder is None:
            raise ValueError("`item_encoder` may not be none")

    def validate_value(self, value):
        if not is_list_like(value):
            raise EncodingTypeError(
                "Cannot encode value of type {0} using array encoder.  Must be "
                "a list-like object such as an array or tuple".format(
                    type(value),
                )
            )

        for item in value:
            self.item_encoder.validate_value(item)

    def encode_elements(self, value):
        self.validate_value(value)

        item_encoder = self.item_encoder
        tail_chunks = tuple(item_encoder(i) for i in value)

        items_are_dynamic = getattr(item_encoder, 'is_dynamic', False)
        if not items_are_dynamic:
            return b''.join(tail_chunks)

        head_length = 32 * len(value)
        tail_offsets = (0,) + tuple(accumulate(map(len, tail_chunks[:-1])))
        head_chunks = tuple(
            encode_uint_256(head_length + offset)
            for offset in tail_offsets
        )
        return b''.join(head_chunks + tail_chunks)

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        item_encoder = registry.get_encoder(str(abi_type.item_type))

        array_spec = abi_type.arrlist[-1]
        if len(array_spec) == 1:
            # If array dimension is fixed
            return SizedArrayEncoder(
                array_size=array_spec[0],
                item_encoder=item_encoder,
            )
        else:
            # If array dimension is dynamic
            return DynamicArrayEncoder(item_encoder=item_encoder)


class SizedArrayEncoder(BaseArrayEncoder):
    array_size = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.is_dynamic = self.item_encoder.is_dynamic

    def validate(self):
        super().validate()

        if self.array_size is None:
            raise ValueError("`array_size` may not be none")

    def validate_value(self, value):
        super().validate_value(value)

        if len(value) != self.array_size:
            raise ValueOutOfBounds(
                "Expected value with length {0}.  Provided value has {1} "
                "elements".format(self.array_size, len(value))
            )

    def encode(self, value):
        encoded_elements = self.encode_elements(value)

        return encoded_elements


class DynamicArrayEncoder(BaseArrayEncoder):
    is_dynamic = True

    def encode(self, value):
        encoded_size = encode_uint_256(len(value))
        encoded_elements = self.encode_elements(value)
        encoded_value = encoded_size + encoded_elements

        return encoded_value



from platone.utils.toolz import (
    assoc,
)


def gas_price_strategy_middleware(make_request, web3):
    """
    Includes a gas price using the gas price strategy
    """
    def middleware(method, params):
        if method == 'platon_sendTransaction':
            transaction = params[0]
            if 'gasPrice' not in transaction:
                generated_gas_price = web3.platone.generateGasPrice(transaction)
                if generated_gas_price is not None:
                    transaction = assoc(transaction, 'gasPrice', generated_gas_price)
                    return make_request(method, [transaction])
        return make_request(method, params)
    return middleware



from platone.packages.eth_utils import (
    to_dict,
)

from platone.utils.abi import (
    map_abi_data,
)
from platone.utils.formatters import (
    apply_formatter_at_index,
)
from platone.utils.toolz import (
    curry,
)

TRANSACTION_PARAMS_ABIS = {
    'data': 'bytes',
    'from': 'address',
    'gas': 'uint',
    'gasPrice': 'uint',
    'nonce': 'uint',
    'to': 'address',
    'value': 'uint',
}

FILTER_PARAMS_ABIS = {
    'to': 'address',
    'address': 'address[]',
}

TRACE_PARAMS_ABIS = {
    'to': 'address',
    'from': 'address',
}

RPC_ABIS = {
    # platon
    # 'platon_call': ['ledger', TRANSACTION_PARAMS_ABIS],
    # 'platon_estimateGas': ['ledger', TRANSACTION_PARAMS_ABIS],
    'platon_getBalance': ['ledger', 'address', None],
    'platon_getBlockByHash': ['ledger', 'bytes32', 'bool'],
    'platon_getBlockTransactionCountByHash': ['ledger', 'bytes32'],
    'platon_getCode': ['ledger', 'address', None],
    'platon_getLogs': ['ledger', FILTER_PARAMS_ABIS],
    'platon_getStorageAt': ['ledger', 'address', 'uint', None],
    'platon_getTransactionByBlockHashAndIndex': ['ledger', 'bytes32', 'uint'],
    'platon_getTransactionByHash': ['ledger', 'bytes32'],
    'platon_getTransactionCount': ['ledger', 'address', None],
    'platon_getTransactionReceipt': ['ledger', 'bytes32'],
    'platon_getUncleCountByBlockHash': ['ledger', 'bytes32'],
    'platon_newFilter': ['ledger', FILTER_PARAMS_ABIS],
    'platon_sendRawTransaction': ['ledger', 'bytes'],
    'platon_sendTransaction': ['ledger', TRANSACTION_PARAMS_ABIS],
    'platon_sign': ['ledger', 'address', 'bytes'],
    # personal
    'personal_sendTransaction': ['ledger', TRANSACTION_PARAMS_ABIS],
    'personal_lockAccount': ['ledger', 'address'],
    'personal_unlockAccount': ['ledger', 'address', None, None],
    'personal_sign': ['ledger', None, 'address', None],
    'trace_call': ['ledger', TRACE_PARAMS_ABIS],
}


@curry
def apply_abi_formatters_to_dict(normalizers, abi_dict, data):
    fields = list(set(abi_dict.keys()) & set(data.keys()))
    formatted_values = map_abi_data(
        normalizers,
        [abi_dict[field] for field in fields],
        [data[field] for field in fields],
    )
    formatted_dict = dict(zip(fields, formatted_values))
    return dict(data, **formatted_dict)


@to_dict
def abi_request_formatters(normalizers, abis):
    for method, abi_types in abis.items():
        if isinstance(abi_types, list):
            yield method, map_abi_data(normalizers, abi_types)
        elif isinstance(abi_types, dict):
            single_dict_formatter = apply_abi_formatters_to_dict(normalizers, abi_types)
            yield method, apply_formatter_at_index(single_dict_formatter, 0)
        else:
            raise TypeError("ABI definitions must be a list or dictionary, got %r" % abi_types)




import copy
import datetime
import functools

from platone.packages.eth_utils import (
    is_same_address,
    remove_0x_prefix,
    to_normalized_address,
)
import idna

from platone.packages.ens.constants import (
    ACCEPTABLE_STALE_HOURS,
    AUCTION_START_GAS_CONSTANT,
    AUCTION_START_GAS_MARGINAL,
    EMPTY_SHA3_BYTES,
    MIN_ETH_LABEL_LENGTH,
    RECOGNIZED_TLDS,
    REVERSE_REGISTRAR_DOMAIN,
)
from platone.packages.ens.exceptions import (
    InvalidLabel,
    InvalidName,
)

default = object()


def Web3():
    from platone import Web3
    return Web3


def dict_copy(func):
    "copy dict keyword args, to avoid modifying caller's copy"
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        copied_kwargs = copy.deepcopy(kwargs)
        return func(*args, **copied_kwargs)
    return wrapper


def ensure_hex(data):
    if not isinstance(data, str):
        return Web3().toHex(data)
    return data


def init_web3(providers=default):
    from platone import Web3

    if providers is default:
        w3 = Web3(ens=None)
    else:
        w3 = Web3(providers, ens=None)

    return customize_web3(w3)


def customize_web3(w3):
    from platone.contract import ConciseContract
    from platone.middleware import make_stalecheck_middleware

    w3.middleware_stack.remove('name_to_address')
    w3.middleware_stack.add(
        make_stalecheck_middleware(ACCEPTABLE_STALE_HOURS * 3600),
        name='stalecheck',
    )
    w3.platone.setContractFactory(ConciseContract)
    return w3


def normalize_name(name):
    '''
    Clean the fully qualified name, as defined in ENS `EIP-137
    <https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax>`_

    This does *not* enforce whether ``name`` is a label or fully qualified domain.

    :param str name: the dot-separated ENS name
    :raises InvalidName: if ``name`` has invalid syntax
    '''
    if not name:
        return name
    elif isinstance(name, (bytes, bytearray)):
        name = name.decode('utf-8')
    try:
        return idna.decode(name, uts46=True, std3_rules=True)
    except idna.IDNAError as exc:
        raise InvalidName("%s is an invalid name, because %s" % (name, exc)) from exc


def is_valid_name(name):
    '''
    Validate whether the fully qualified name is valid, as defined in ENS `EIP-137
    <https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax>`_

    :param str name: the dot-separated ENS name
    :returns: True if ``name`` is set, and :meth:`~ens.main.ENS.nameprep` will not raise InvalidName
    '''
    if not name:
        return False
    try:
        normalize_name(name)
        return True
    except InvalidName:
        return False


def label_to_name(label, default_tld, recognized_tlds):
    label = normalize_name(label)
    pieces = label.split('.')
    if pieces[-1] not in recognized_tlds:
        pieces.append(default_tld)
    return '.'.join(pieces)


def dot_eth_name(label):
    return label_to_name(label, 'eth', RECOGNIZED_TLDS)


def name_to_label(name, registrar):
    name = normalize_name(name)
    if '.' not in name:
        label = name
    else:
        name_pieces = name.split('.')
        registrar_pieces = registrar.split('.')
        if len(name_pieces) != len(registrar_pieces) + 1:
            raise ValueError(
                "You must specify a label, like 'tickets' "
                "or a fully-qualified name, like 'tickets.%s'" % registrar
            )
        label, *label_registrar = name_pieces
        if label_registrar != registrar_pieces:
            raise ValueError("This interface only manages names under .%s " % registrar)
    return label


def dot_eth_label(name):
    '''
    Convert from a name, like 'ethfinex.eth', to a label, like 'ethfinex'
    If name is already a label, this should be a noop, except for converting to a string
    and validating the name syntax.
    '''
    label = name_to_label(name, registrar='eth')
    if len(label) < MIN_ETH_LABEL_LENGTH:
        raise InvalidLabel('name %r is too short' % label)
    else:
        return label


def to_utc_datetime(timestamp):
    if timestamp:
        return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)
    else:
        return None


def sha3_text(val):
    if isinstance(val, str):
        val = val.encode('utf-8')
    return Web3().sha3(val)


def label_to_hash(label):
    label = normalize_name(label)
    if '.' in label:
        raise ValueError("Cannot generate hash for label %r with a '.'" % label)
    return Web3().sha3(text=label)


def name_to_hash(name):
    node = EMPTY_SHA3_BYTES
    if name:
        labels = name.split(".")
        for label in reversed(labels):
            labelhash = label_to_hash(label)
            assert isinstance(labelhash, bytes)
            assert isinstance(node, bytes)
            node = Web3().sha3(node + labelhash)
    return node


def dot_eth_namehash(name):
    '''
    Generate the namehash. This is also known as the ``node`` in ENS contracts.

    In normal operation, generating the namehash is handled
    behind the scenes. For advanced usage, it is a helpful utility.

    This will add '.eth' to name if no TLD given. Also, it normalizes the name with
    `nameprep
    <https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax>`_
    before hashing.

    :param str name: ENS name to hash
    :return: the namehash
    :rtype: bytes
    :raises InvalidName: if ``name`` has invalid syntax
    '''
    expanded_name = dot_eth_name(name)
    return name_to_hash(expanded_name)


def address_in(address, addresses):
    return any(is_same_address(address, item) for item in addresses)


def address_to_reverse_domain(address):
    lower_unprefixed_address = remove_0x_prefix(to_normalized_address(address))
    return lower_unprefixed_address + '.' + REVERSE_REGISTRAR_DOMAIN


def estimate_auction_start_gas(labels):
    return AUCTION_START_GAS_CONSTANT + AUCTION_START_GAS_MARGINAL * len(labels)


def assert_signer_in_modifier_kwargs(modifier_kwargs):
    ERR_MSG = "You must specify the sending account"
    assert len(modifier_kwargs) == 1, ERR_MSG

    _modifier_type, modifier_dict = dict(modifier_kwargs).popitem()
    if 'from' not in modifier_dict:
        raise TypeError(ERR_MSG)

    return modifier_dict['from']



from platone.utils.fvnhash import fnv1_64


def hexstr2bytes(address: str):
    pos = 0
    len_str = len(address)
    if len_str % 2 != 0:
        return None
    len_str = round(len_str / 2)
    hexa = []
    for i in range(len_str):
        s1 = address[pos:pos + 2]
        if s1 == '0x' or s1 == '0X':
            pos += 2
            continue
        sv = s1
        hexa.append(sv)
        pos += 2
    return hexa


def tostring_hex(arr: list):
    arrhex = ''
    if arr:
        for i in arr:
            arrhex = arrhex + i
        return arrhex
    else:
        return ''


def stringtohex(str1: bytes):
    strhex = []
    if str1:
        for i in str1:
            temp = hex(i).replace('0x', '')
            if len(temp) == 1:
                temp = '0' + temp
            strhex = strhex + [temp]
            del temp
        return strhex
    else:
        return []


def stringfnv(str1: str):
    a1 = bytes(str1, 'utf-8')
    a2 = fnv1_64(a1)
    a3 = hex(a2)
    if len(a3) % 2:
        a4 = '0' + a3.replace('0x', '')
        aa = hexstr2bytes(a4)
    else:
        aa = hexstr2bytes(a3)
    return aa


def rlp_encode(input):
    if isinstance(input, list):
        if len(input) == 1 and input[0] == '0':
            return ['80']
        elif len(input) == 1 and int(input[0], 16) <= 0x7f:
            return input
        elif len(input) == 0:
            return []
        else:
            temp = encode_length(len(input), 0x80)
            return (hexstr2bytes(temp.replace('0x', '')) + input)
    elif isinstance(input, tuple):
        output = []
        for item in input: output += rlp_encode(item)
        temp1 = encode_length(len(output), 0xc0)
        return hexstr2bytes(temp1.replace('0x', '')) + output


def encode_length(L, offset):
    if L < 56:
        return hex(L + offset)
    elif L < 256 ** 8:
        BL = to_binary(L)
        return hex(int(len(BL) / 2) + offset + 55) + BL
    else:
        raise Exception("input too long")


def to_binary(x):
    if x == 0:
        return ''
    else:
        tem1 = (to_binary(int(x / 256))).replace('0x', '')
        tem2 = hex(x % 256).replace('0x', '')
        if tem1:
            if len(tem1) == 1:
                tem1 = '0' + tem1
        if tem2:
            if len(tem2) == 1:
                tem2 = '0' + tem2
        return tem1 + tem2
    # if x == 0:
    #     return ''
    # else: return hex(x)


def listcat(list1, list2):
    if list2:
        if isinstance(list1[0], list):
            if isinstance(list2[0], list):
                for i in range(len(list2)):
                    list1.append(list2[i])
                temp = list1
            else:
                temp = [list1, list2]
        else:
            temp = [list1]
            if isinstance(list2[0], list):
                for i in range(len(list2)):
                    temp.append(list2[i])
            else:
                temp.append(list2)
    else:
        temp = list1
    return temp


def tuplecat(tuple1, tuple2):
    # if isinstance(tuple1[0],tuple):
    #     if isinstance(tuple2[0],tuple):
    #         temp = (tuple1,tuple2)
    #     else:
    #         temp1 = list(tuple1)
    #         temp1.append(tuple2)
    #         temp = tuple(temp1)
    # elif tuple2:
    #     if isinstance(tuple2[0],tuple):
    #         temp1 = list(tuple2)
    #         temp1.insert(tuple1)
    #         temp = tuple(temp1)
    #     elif isinstance(tuple2,tuple):
    #         temp = (tuple1,tuple2)
    #     else:
    #         if isinstance(tuple2[0],list):
    #             temp1 = list(tuple1)
    #             for i in range(len(tuple2)):
    #                 temp1.append(tuple2[i])
    #             temp = tuple(temp1)
    #         else:
    #             temp1 = list(tuple1)
    #             temp1.append(tuple2)
    #             temp = tuple(temp1)
    if tuple2:
        if isinstance(tuple2, tuple):
            temp = (tuple1, tuple2)
        elif isinstance(tuple2, list):
            if isinstance(tuple2[0], list):
                temp = [tuple1]
                for i in range(len(tuple2)):
                    temp.append(tuple2[i])
                temp = tuple(temp)
            else:
                temp = (tuple1, tuple2)
    else:
        temp = tuple1

    return temp


def rlp_decode(input):
    if len(input) == 0:
        return []
    output = []
    (offset, dataLen, type) = decode_length(input)
    if type is list:
        output = input[offset:offset + dataLen]
        output = listcat(output, rlp_decode(input[offset + dataLen:]))
    elif type is tuple:
        output = rlp_decode(input[offset:offset + dataLen])
        if input[offset + dataLen:]:
            if isinstance(output, list):
                output = tuple([output])
            output = tuplecat(output, rlp_decode(input[offset + dataLen:]))
        else:
            if isinstance(output, list):
                output = tuple([output])
    return output


def decode_length(input):
    length = len(input)
    if length == 0:
        raise Exception("input is null")
    prefix = int(input[0], 16)
    if prefix <= 0x7f:
        return (0, 1, list)
    elif prefix <= 0xb7 and length > prefix - 0x80:
        strLen = prefix - 0x80
        return (1, strLen, list)
    elif prefix <= 0xbf and length > prefix - 0xb7 and length > prefix - 0xb7 + to_integer(input[1:prefix - 0xb6]):
        lenOfStrLen = prefix - 0xb7
        strLen = to_integer(input[1:lenOfStrLen + 1])
        return (1 + lenOfStrLen, strLen, list)
    elif prefix <= 0xf7 and length > prefix - 0xc0:
        listLen = prefix - 0xc0
        return (1, listLen, tuple)
    elif prefix <= 0xff and length > prefix - 0xf7 and length > prefix - 0xf7 + to_integer(input[1:prefix - 0xf6]):
        lenOfListLen = prefix - 0xf7
        listLen = to_integer(input[1:lenOfListLen + 1])
        return (1 + lenOfListLen, listLen, tuple)
    else:
        raise Exception("input don't conform RLP encoding form")


def to_integer(b):
    # 获得list或tuple的真实长度
    length = len(b)
    if length == 0:
        raise Exception("input is null")
    elif length == 1:
        return int(b[0], 16)
    else:
        return int(b[-1:], 16) + to_integer(b[0:-1]) * 256



import json
from platone.module import Module
from platone.utils.builtin_abis import USER_MANAGER_ABI
from platone.utils.transactions import send_obj_transaction


class User(Module):
    ABI = USER_MANAGER_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.user_manager_address, abi=self.ABI,
                                                   ledger=self.web3.system_ledger)

    def getByAddress(self, address):
        return self.contract.functions.GetByAddr(address).call()

    def getByName(self, name):
        return self.contract.functions.GetByName(name).call()

    def hasRoles(self, address, roles):
        return self.contract.functions.HasRoles(address, roles).call()

    def pageAll(self, page_num, page_size):
        return self.contract.functions.PageAll(page_num, page_size).call()

    def pageByRoles(self, roles, page_num, page_size):
        return self.contract.functions.PageByRoles(roles, page_num, page_size).call()

    def pageByStat(self, status, page_num, page_size):
        return self.contract.functions.PageByStat(status, page_num, page_size).call()

    def add(self, address, name, mobile, email, desc, roles, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Add(address, name, mobile, email, desc, roles).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def update(self, address, mobile, email, desc, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Update(address, mobile, email, desc).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def delete(self, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Delete(address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def addRoles(self, address, roles, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.AddRoles(address, roles).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def setRoles(self, address, roles, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.SetRoles(address, roles).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def removeRoles(self, address, roles, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.RemoveRoles(address, roles).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def audit(self, address, audit_stat, audit_reason, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Audit(address, audit_stat, audit_reason).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def register(self, name, mobile, email, desc, roles, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Register(name, mobile, email, desc, roles).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def enable(self, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Enable(address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def disable(self, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Disable(address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)



from platone.packages.eth_utils import (
    is_string,
)


class VersionModuleTest:
    def test_net_version(self, web3):
        version = web3.version.network

        assert is_string(version)
        assert version.isdigit()

    def test_eth_protocolVersion(self, web3):
        protocol_version = web3.version.ethereum

        assert is_string(protocol_version)
        assert protocol_version.isdigit()



def rlp_encode(input):
    if isinstance(input,list):
        if len(input) == 1 and ord(input[0]) <= 0x7f: return input
        else:
            temp=encode_length(len(input), 0x80)
            return ([temp.replace('0x','')]+input)
    elif isinstance(input,tuple):
        output = []
        for item in input: output += rlp_encode(item)
        return encode_length(len(output), 0xc0) + output

def encode_length(L,offset):
    if L < 56:
         return hex(L + offset)
    elif L < 256**8:
         BL = to_binary(L)
         return hex(len(BL) + offset + 55) + BL
    else:
         raise Exception("input too long")

def to_binary(x):
    if x == 0:
        return ''
    else:
        return to_binary(int(x / 256)) + hex(x % 256)


# def rlp_decode(input):
#     if len(input) == 0:
#         return
#     output = ''
#     (offset, dataLen, type) = decode_length(input)
#     if type is str:
#         output = instantiate_str(substr(input, offset, dataLen))
#     elif type is list:
#         output = instantiate_list(substr(input, offset, dataLen))
#     output + rlp_decode(substr(input, offset + dataLen))
#     return output
#
# def decode_length(input):
#     length = len(input)
#     if length == 0:
#         raise Exception("input is null")
#     prefix = ord(input[0])
#     if prefix <= 0x7f:
#         return (0, 1, str)
#     elif prefix <= 0xb7 and length > prefix - 0x80:
#         strLen = prefix - 0x80
#         return (1, strLen, str)
#     elif prefix <= 0xbf and length > prefix - 0xb7 and length > prefix - 0xb7 + to_integer(substr(input, 1, prefix - 0xb7)):
#         lenOfStrLen = prefix - 0xb7
#         strLen = to_integer(substr(input, 1, lenOfStrLen))
#         return (1 + lenOfStrLen, strLen, str)
#     elif prefix <= 0xf7 and length > prefix - 0xc0:
#         listLen = prefix - 0xc0;
#         return (1, listLen, list)
#     elif prefix <= 0xff and length > prefix - 0xf7 and length > prefix - 0xf7 + to_integer(substr(input, 1, prefix - 0xf7)):
#         lenOfListLen = prefix - 0xf7
#         listLen = to_integer(substr(input, 1, lenOfListLen))
#         return (1 + lenOfListLen, listLen, list)
#     else:
#         raise Exception("input don't conform RLP encoding form")
#
# def to_integer(b):
#     length = len(b)
#     if length == 0:
#         raise Exception("input is null")
#     elif length == 1:
#         return ord(b[0])
#     else:
#         return ord(substr(b, -1)) + to_integer(substr(b, 0, -1)) * 256



from __future__ import absolute_import

import os
from typing import Type

from platone.packages.platone_keys.utils.module_loading import (
    import_string,
)

from .base import BaseECCBackend  # noqa: F401
from .coincurve import (  # noqa: F401
    CoinCurveECCBackend,
    is_coincurve_available,
)
from .native import NativeECCBackend  # noqa: F401


def get_default_backend_class() -> str:
    if is_coincurve_available():
        return 'platone.packages.platone_keys.backends.CoinCurveECCBackend'
    else:
        return 'platone.packages.platone_keys.backends.NativeECCBackend'


def get_backend_class(import_path: str = None) -> Type[BaseECCBackend]:
    if import_path is None:
        import_path = os.environ.get(
            'ECC_BACKEND_CLASS',
            get_default_backend_class(),
        )
    return import_string(import_path)


def get_backend(import_path: str = None) -> BaseECCBackend:
    backend_class = get_backend_class(import_path)
    return backend_class()



import json
from platone.module import Module
from platone.utils.builtin_abis import LEDGER_MANAGER_ABI
from platone.utils.transactions import send_obj_transaction


class Ledger(Module):
    ABI = LEDGER_MANAGER_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.ledger_manager_address, abi=self.ABI,
                                                   ledger=self.web3.system_ledger)

    def getAllLedgers(self):
        return self.contract.functions.GetAllLedgers().call()

    def createLedger(self, ledger_json, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.CreateLedger(json.dumps(ledger_json)).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def joinLedger(self, ledger_name, node_id, bls_pubKey, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.JoinLedger(ledger_name, node_id, bls_pubKey).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def joinLedgers(self, node_id, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.JoinedLedgers(node_id).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def quitLedger(self, ledger_name, node_id, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.QuitLedger(ledger_name, node_id).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def terminateLedger(self, ledger_name, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.TerminateLedger(ledger_name).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)



import json
from platone.module import Module
from platone.utils.builtin_abis import LEDGER_NODE_MANAGER_ABI
from platone.utils.transactions import send_obj_transaction


class LedgerNode(Module):
    ABI = LEDGER_NODE_MANAGER_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.ledger_node_manager_address, abi=self.ABI,
                                                   ledger=self.web3.system_ledger)

    def list(self):
        return self.contract.functions.List().call()

    def add(self, pubkey, bls_pubkey, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Add(pubkey, bls_pubkey).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def grantPerm(self, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.GrantPerm(address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def remove(self, pubkey, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Remove(pubkey).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def revokePerm(self, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.RevokePerm(address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def updateNodeType(self, pubkey, node_type, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.UpdateNodeType(pubkey, node_type).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)




from typing import Any, Iterable, Iterator, Tuple, Union
from urllib import parse

from eth_typing import URI, Hash32

from .toolz import sliding_window, take


def humanize_seconds(seconds: Union[float, int]) -> str:
    if int(seconds) == 0:
        return "0s"

    unit_values = _consume_leading_zero_units(_humanize_seconds(int(seconds)))

    return "".join(
        ("{0}{1}".format(amount, unit) for amount, unit in take(3, unit_values))
    )


SECOND = 1
MINUTE = 60
HOUR = 60 * 60
DAY = 24 * HOUR
YEAR = 365 * DAY
MONTH = YEAR // 12
WEEK = 7 * DAY


UNITS = (
    (YEAR, "y"),
    (MONTH, "m"),
    (WEEK, "w"),
    (DAY, "d"),
    (HOUR, "h"),
    (MINUTE, "m"),
    (SECOND, "s"),
)


def _consume_leading_zero_units(
    units_iter: Iterator[Tuple[int, str]]
) -> Iterator[Tuple[int, str]]:
    for amount, unit in units_iter:
        if amount == 0:
            continue
        else:
            yield (amount, unit)
            break

    yield from units_iter


def _humanize_seconds(seconds: int) -> Iterator[Tuple[int, str]]:
    remainder = seconds

    for duration, unit in UNITS:
        if not remainder:
            break

        num = remainder // duration
        yield num, unit

        remainder %= duration


DISPLAY_HASH_CHARS = 4


def humanize_bytes(value: bytes) -> str:
    if len(value) <= DISPLAY_HASH_CHARS + 1:
        return value.hex()
    value_as_hex = value.hex()
    head = value_as_hex[:DISPLAY_HASH_CHARS]
    tail = value_as_hex[-1 * DISPLAY_HASH_CHARS :]
    return "{0}..{1}".format(head, tail)


def humanize_hash(value: Hash32) -> str:
    return humanize_bytes(value)


def humanize_ipfs_uri(uri: URI) -> str:
    if not is_ipfs_uri(uri):
        raise TypeError(
            "%s does not look like a valid IPFS uri. Currently, "
            "only CIDv0 hash schemes are supported." % uri
        )

    parsed = parse.urlparse(uri)
    ipfs_hash = parsed.netloc
    head = ipfs_hash[:DISPLAY_HASH_CHARS]
    tail = ipfs_hash[-1 * DISPLAY_HASH_CHARS :]
    return "ipfs://{0}..{1}".format(head, tail)


def is_ipfs_uri(value: Any) -> bool:
    if not isinstance(value, str):
        return False

    parsed = parse.urlparse(value)
    if parsed.scheme != "ipfs" or not parsed.netloc:
        return False

    return _is_CIDv0_ipfs_hash(parsed.netloc)


def _is_CIDv0_ipfs_hash(ipfs_hash: str) -> bool:
    if ipfs_hash.startswith("Qm") and len(ipfs_hash) == 46:
        return True
    return False


def _find_breakpoints(*values: int) -> Iterator[int]:
    yield 0
    for index, (left, right) in enumerate(sliding_window(2, values), 1):
        if left + 1 == right:
            continue
        else:
            yield index
    yield len(values)


def _extract_integer_ranges(*values: int) -> Iterator[Tuple[int, int]]:
    """
    Take a sequence of integers which is expected to be ordered and return the
    most concise definition of the sequence in terms of integer ranges.

    - fn(1, 2, 3) -> ((1, 3),)
    - fn(1, 2, 3, 7, 8, 9) -> ((1, 3), (7, 9))
    - fn(1, 7, 8, 9) -> ((1, 1), (7, 9))
    """
    for left, right in sliding_window(2, _find_breakpoints(*values)):
        chunk = values[left:right]
        yield chunk[0], chunk[-1]


def _humanize_range(bounds: Tuple[int, int]) -> str:
    left, right = bounds
    if left == right:
        return str(left)
    else:
        return "{left}-{right}".format(left=left, right=right)


def humanize_integer_sequence(values_iter: Iterable[int]) -> str:
    """
    Return a human readable string that attempts to concisely define a sequence
    of integers.

    - fn((1, 2, 3)) -> '1-3'
    - fn((1, 2, 3, 7, 8, 9)) -> '1-3|7-9'
    - fn((1, 2, 3, 5, 7, 8, 9)) -> '1-3|5|7-9'
    - fn((1, 7, 8, 9)) -> '1|7-9'
    """
    values = tuple(values_iter)
    if not values:
        return "(empty)"
    else:
        return "|".join(map(_humanize_range, _extract_integer_ranges(*values)))



from typing import List


class Module:
    web3 = None

    def __init__(self, web3):
        self.web3 = web3

    @classmethod
    def attach(cls, target, module_name=None):
        if not module_name:
            module_name = cls.__name__.lower()

        if hasattr(target, module_name):
            raise AttributeError(
                "Cannot set {0} module named '{1}'.  The web3 object "
                "already has an attribute with that name".format(
                    target,
                    module_name,
                )
            )

        if isinstance(target, Module):
            web3 = target.web3
        else:
            web3 = target

        setattr(target, module_name, cls(web3))

    def set_ledger(self, args: List, ledger, index=0):
        if ledger is None:
            if self.web3.currentLedger is None:
                return args
            ledger = self.web3.currentLedger
        args.insert(index, ledger)
        return args



from platone.module import (
    Module,
)


class Miner(Module):

    # todo： coding
    def setGasPrice(self, gas_price, ledger=None):
        return self.web3.manager.request_blocking(
            "miner_setGasPrice", self.set_ledger([gas_price], ledger),
        )


from platone.packages.platone_account.signers.base import (
    BaseAccount,
)
from platone.packages.platone_keys.utils.address import MIANNETHRP, TESTNETHRP
from platone.packages.platone_keys.datatypes import sm3_tobech32_address,sm3_tobech32_testaddress


class LocalAccount(BaseAccount):
    '''
    A collection of convenience methods to sign and encrypt, with an embedded private key.

    :var bytes privateKey: the 32-byte private key data

    .. code-block:: python

        >>> my_local_account.address
        "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55"
        >>> my_local_account.privateKey
        b"\\x01\\x23..."

    You can also get the private key by casting the account to :class:`bytes`:

    .. code-block:: python

        >>> bytes(my_local_account)
        b"\\x01\\x23..."
    '''

    def __init__(self, key, account, net_type=MIANNETHRP, mode='ECDSA'):
        '''
        :param platone_keys.PrivateKey key: to prefill in private key execution
        :param web3.account.Account account: the key-unaware management API
        '''
        self._publicapi = account

        if mode == 'SM':
            addr_dict = {MIANNETHRP: sm3_tobech32_address(key.public_key),
                         TESTNETHRP: sm3_tobech32_testaddress(key.public_key)}
        else:
            addr_dict = {MIANNETHRP: key.public_key.to_bech32_address(),
                         TESTNETHRP: key.public_key.to_bech32_test_address()}

        self._address = addr_dict[net_type]

        key_raw = key.to_bytes()
        self._privateKey = key_raw

        self._key_obj = key

    @property
    def address(self):
        return self._address

    @property
    def privateKey(self):
        '''
        Get the private key.
        '''
        return self._privateKey

    def encrypt(self, password):
        '''
        Generate a string with the encrypted key, as in
        :meth:`~platone_account.account.Account.encrypt`, but without a private key argument.
        '''
        return self._publicapi.encrypt(self.privateKey, password)

    def signHash(self, message_hash):
        return self._publicapi.signHash(
            message_hash,
            private_key=self.privateKey,
        )

    def signTransaction(self, transaction_dict):
        return self._publicapi.signTransaction(transaction_dict, self.privateKey)

    def __bytes__(self):
        return self.privateKey




import json
from platone.module import Module
from platone.utils.builtin_abis import NODE_MANAGER_ABI
from platone.utils.transactions import send_obj_transaction


class Node(Module):
    ABI = NODE_MANAGER_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.node_manager_address, abi=self.ABI,
                                                   ledger=self.web3.system_ledger)

    def listAll(self):
        return self.contract.functions.ListAll().call()

    def listByHostAddress(self, host_address):
        return self.contract.functions.ListByHostAddress(host_address).call()

    def listByOwner(self, owner):
        return self.contract.functions.ListByOwner(owner).call()

    def listByStat(self, status):
        return self.contract.functions.ListByStat(status).call()

    def listByType(self, node_type):
        return self.contract.functions.ListByType(node_type).call()

    def getByName(self, name):
        return self.contract.functions.GetByName(name).call()

    def getByPublicKey(self, pubkey):
        return self.contract.functions.GetByPublicKey(pubkey).call()

    def add(self, name, owner, desc, pubkey, bls_pubkey, host_address, rpc_port, p2p_port, private_key,
            tx_cfg: dict = None):
        txn = self.contract.functions.Add(name, owner, desc, pubkey, bls_pubkey, host_address, rpc_port,
                                          p2p_port).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def delete(self, name, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Delete(name).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def apply(self, name, desc, pubkey, bls_pubkey, host_address, rpc_port, p2p_port, private_key,
              tx_cfg: dict = None):
        txn = self.contract.functions.Apply(name, desc, pubkey, bls_pubkey, host_address, rpc_port,
                                            p2p_port).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def audit(self, name, audit_stat, audit_reason, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Audit(name, audit_stat, audit_reason).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def update(self, name, host_address, rpc_port, p2p_port, desc, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Update(name, host_address, rpc_port, p2p_port, desc).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def updateType(self, name, node_type, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.UpdateType(name, node_type).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def enable(self, name, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Enable(name).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def disable(self, name, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Disable(name).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)



# -*- coding: utf-8 -*-

import pytest

from platone.packages.eth_abi import (
    decode_single,
)
from platone.packages.eth_utils import (
    is_boolean,
    is_bytes,
    is_checksum_address,
    is_dict,
    is_integer,
    is_list_like,
    is_same_address,
    is_string,
)
from hexbytes import (
    HexBytes,
)

from platone.exceptions import (
    InvalidAddress,
)

UNKNOWN_ADDRESS = '0xdEADBEeF00000000000000000000000000000000'
UNKNOWN_HASH = '0xdeadbeef00000000000000000000000000000000000000000000000000000000'


class EthModuleTest:
    def test_eth_protocolVersion(self, web3):
        protocol_version = web3.version.ethereum

        assert is_string(protocol_version)
        assert protocol_version.isdigit()

    def test_eth_syncing(self, web3):
        syncing = web3.platone.syncing

        assert is_boolean(syncing) or is_dict(syncing)

        if is_boolean(syncing):
            assert syncing is False
        elif is_dict(syncing):
            assert 'startingBlock' in syncing
            assert 'currentBlock' in syncing
            assert 'highestBlock' in syncing

            assert is_integer(syncing['startingBlock'])
            assert is_integer(syncing['currentBlock'])
            assert is_integer(syncing['highestBlock'])

    def test_eth_coinbase(self, web3):
        coinbase = web3.platone.coinbase
        assert is_checksum_address(coinbase)

    def test_eth_mining(self, web3):
        mining = web3.platone.mining
        assert is_boolean(mining)

    def test_eth_hashrate(self, web3):
        hashrate = web3.platone.hashrate
        assert is_integer(hashrate)
        assert hashrate >= 0

    def test_eth_gasPrice(self, web3):
        gas_price = web3.platone.gasPrice
        assert is_integer(gas_price)
        assert gas_price > 0

    def test_eth_accounts(self, web3):
        accounts = web3.platone.accounts
        assert is_list_like(accounts)
        assert len(accounts) != 0
        assert all((
            is_checksum_address(account)
            for account
            in accounts
        ))
        assert web3.platone.coinbase in accounts

    def test_eth_blockNumber(self, web3):
        block_number = web3.platone.blockNumber
        assert is_integer(block_number)
        assert block_number >= 0

    def test_eth_getBalance(self, web3):
        coinbase = web3.platone.coinbase

        with pytest.raises(InvalidAddress):
            web3.platone.getBalance(coinbase.lower())

        balance = web3.platone.getBalance(coinbase)

        assert is_integer(balance)
        assert balance >= 0

    def test_eth_getStorageAt(self, web3, emitter_contract_address):
        storage = web3.platone.getStorageAt(emitter_contract_address, 0)
        assert isinstance(storage, HexBytes)

    def test_eth_getStorageAt_invalid_address(self, web3):
        coinbase = web3.platone.coinbase
        with pytest.raises(InvalidAddress):
            web3.platone.getStorageAt(coinbase.lower(), 0)

    def test_eth_getTransactionCount(self, web3, unlocked_account_dual_type):
        transaction_count = web3.platone.getTransactionCount(unlocked_account_dual_type)
        assert is_integer(transaction_count)
        assert transaction_count >= 0

    def test_eth_getTransactionCount_invalid_address(self, web3):
        coinbase = web3.platone.coinbase
        with pytest.raises(InvalidAddress):
            web3.platone.getTransactionCount(coinbase.lower())

    def test_eth_getBlockTransactionCountByHash_empty_block(self, web3, empty_block):
        transaction_count = web3.platone.getBlockTransactionCount(empty_block['hash'])

        assert is_integer(transaction_count)
        assert transaction_count == 0

    def test_eth_getBlockTransactionCountByNumber_empty_block(self, web3, empty_block):
        transaction_count = web3.platone.getBlockTransactionCount(empty_block['number'])

        assert is_integer(transaction_count)
        assert transaction_count == 0

    def test_eth_getBlockTransactionCountByHash_block_with_txn(self, web3, block_with_txn):
        transaction_count = web3.platone.getBlockTransactionCount(block_with_txn['hash'])

        assert is_integer(transaction_count)
        assert transaction_count >= 1

    def test_eth_getBlockTransactionCountByNumber_block_with_txn(self, web3, block_with_txn):
        transaction_count = web3.platone.getBlockTransactionCount(block_with_txn['number'])

        assert is_integer(transaction_count)
        assert transaction_count >= 1

    def test_eth_getUncleCountByBlockHash(self, web3, empty_block):
        uncle_count = web3.platone.getUncleCount(empty_block['hash'])

        assert is_integer(uncle_count)
        assert uncle_count == 0

    def test_eth_getUncleCountByBlockNumber(self, web3, empty_block):
        uncle_count = web3.platone.getUncleCount(empty_block['number'])

        assert is_integer(uncle_count)
        assert uncle_count == 0

    def test_eth_getCode(self, web3, math_contract_address):
        code = web3.platone.getCode(math_contract_address)
        assert isinstance(code, HexBytes)
        assert len(code) > 0

    def test_eth_getCode_invalid_address(self, web3, math_contract):
        with pytest.raises(InvalidAddress):
            web3.platone.getCode(math_contract.address.lower())

    def test_eth_getCode_with_block_identifier(self, web3, emitter_contract):
        code = web3.platone.getCode(emitter_contract.address, block_identifier=web3.platone.blockNumber)
        assert isinstance(code, HexBytes)
        assert len(code) > 0

    def test_eth_sign(self, web3, unlocked_account_dual_type):
        signature = web3.platone.sign(
            unlocked_account_dual_type, text='Message tö sign. Longer than hash!'
        )
        assert is_bytes(signature)
        assert len(signature) == 32 + 32 + 1

        # test other formats
        hexsign = web3.platone.sign(
            unlocked_account_dual_type,
            hexstr='0x4d6573736167652074c3b6207369676e2e204c6f6e676572207468616e206861736821'
        )
        assert hexsign == signature

        intsign = web3.platone.sign(
            unlocked_account_dual_type,
            0x4d6573736167652074c3b6207369676e2e204c6f6e676572207468616e206861736821
        )
        assert intsign == signature

        bytessign = web3.platone.sign(
            unlocked_account_dual_type, b'Message t\xc3\xb6 sign. Longer than hash!'
        )
        assert bytessign == signature

        new_signature = web3.platone.sign(
            unlocked_account_dual_type, text='different message is different'
        )
        assert new_signature != signature

    def test_eth_sendTransaction_addr_checksum_required(self, web3, unlocked_account):
        non_checksum_addr = unlocked_account.lower()
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }

        with pytest.raises(InvalidAddress):
            invalid_params = dict(txn_params, **{'from': non_checksum_addr})
            web3.platone.sendTransaction(invalid_params)

        with pytest.raises(InvalidAddress):
            invalid_params = dict(txn_params, **{'to': non_checksum_addr})
            web3.platone.sendTransaction(invalid_params)

    def test_eth_sendTransaction(self, web3, unlocked_account_dual_type):
        txn_params = {
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)
        txn = web3.platone.getTransaction(txn_hash)

        assert is_same_address(txn['from'], txn_params['from'])
        assert is_same_address(txn['to'], txn_params['to'])
        assert txn['value'] == 1
        assert txn['gas'] == 21000
        assert txn['gasPrice'] == txn_params['gasPrice']

    def test_eth_sendTransaction_with_nonce(self, web3, unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            # Increased gas price to ensure transaction hash different from other tests
            'gasPrice': web3.platone.gasPrice * 2,
            'nonce': web3.platone.getTransactionCount(unlocked_account),
        }
        txn_hash = web3.platone.sendTransaction(txn_params)
        txn = web3.platone.getTransaction(txn_hash)

        assert is_same_address(txn['from'], txn_params['from'])
        assert is_same_address(txn['to'], txn_params['to'])
        assert txn['value'] == 1
        assert txn['gas'] == 21000
        assert txn['gasPrice'] == txn_params['gasPrice']
        assert txn['nonce'] == txn_params['nonce']

    def test_eth_replaceTransaction(self, web3, unlocked_account_dual_type):
        txn_params = {
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        txn_params['gasPrice'] = web3.platone.gasPrice * 2
        replace_txn_hash = web3.platone.replaceTransaction(txn_hash, txn_params)
        replace_txn = web3.platone.getTransaction(replace_txn_hash)

        assert is_same_address(replace_txn['from'], txn_params['from'])
        assert is_same_address(replace_txn['to'], txn_params['to'])
        assert replace_txn['value'] == 1
        assert replace_txn['gas'] == 21000
        assert replace_txn['gasPrice'] == txn_params['gasPrice']

    def test_eth_replaceTransaction_non_existing_transaction(
            self, web3, unlocked_account_dual_type):
        txn_params = {
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        with pytest.raises(ValueError):
            web3.platone.replaceTransaction(
                '0x98e8cc09b311583c5079fa600f6c2a3bea8611af168c52e4b60b5b243a441997',
                txn_params
            )

    # auto mine is enabled for this test
    def test_eth_replaceTransaction_already_mined(self, web3, unlocked_account_dual_type):
        txn_params = {
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        txn_params['gasPrice'] = web3.platone.gasPrice * 2
        with pytest.raises(ValueError):
            web3.platone.replaceTransaction(txn_hash, txn_params)

    def test_eth_replaceTransaction_incorrect_nonce(self, web3, unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)
        txn = web3.platone.getTransaction(txn_hash)

        txn_params['gasPrice'] = web3.platone.gasPrice * 2
        txn_params['nonce'] = txn['nonce'] + 1
        with pytest.raises(ValueError):
            web3.platone.replaceTransaction(txn_hash, txn_params)

    def test_eth_replaceTransaction_gas_price_too_low(self, web3, unlocked_account_dual_type):
        txn_params = {
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': 10,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        txn_params['gasPrice'] = 9
        with pytest.raises(ValueError):
            web3.platone.replaceTransaction(txn_hash, txn_params)

    def test_eth_replaceTransaction_gas_price_defaulting_minimum(self, web3, unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': 10,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        txn_params.pop('gasPrice')
        replace_txn_hash = web3.platone.replaceTransaction(txn_hash, txn_params)
        replace_txn = web3.platone.getTransaction(replace_txn_hash)

        assert replace_txn['gasPrice'] == 11  # minimum gas price

    def test_eth_replaceTransaction_gas_price_defaulting_strategy_higher(self,
                                                                         web3,
                                                                         unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': 10,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        def higher_gas_price_strategy(web3, txn):
            return 20

        web3.platone.setGasPriceStrategy(higher_gas_price_strategy)

        txn_params.pop('gasPrice')
        replace_txn_hash = web3.platone.replaceTransaction(txn_hash, txn_params)
        replace_txn = web3.platone.getTransaction(replace_txn_hash)
        assert replace_txn['gasPrice'] == 20  # Strategy provides higher gas price

    def test_eth_replaceTransaction_gas_price_defaulting_strategy_lower(self,
                                                                        web3,
                                                                        unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': 10,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        def lower_gas_price_strategy(web3, txn):
            return 5

        web3.platone.setGasPriceStrategy(lower_gas_price_strategy)

        txn_params.pop('gasPrice')
        replace_txn_hash = web3.platone.replaceTransaction(txn_hash, txn_params)
        replace_txn = web3.platone.getTransaction(replace_txn_hash)
        # Strategy provices lower gas price - minimum preferred
        assert replace_txn['gasPrice'] == 11

    def test_eth_modifyTransaction(self, web3, unlocked_account):
        txn_params = {
            'from': unlocked_account,
            'to': unlocked_account,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        }
        txn_hash = web3.platone.sendTransaction(txn_params)

        modified_txn_hash = web3.platone.modifyTransaction(
            txn_hash, gasPrice=(txn_params['gasPrice'] * 2), value=2
        )
        modified_txn = web3.platone.getTransaction(modified_txn_hash)

        assert is_same_address(modified_txn['from'], txn_params['from'])
        assert is_same_address(modified_txn['to'], txn_params['to'])
        assert modified_txn['value'] == 2
        assert modified_txn['gas'] == 21000
        assert modified_txn['gasPrice'] == txn_params['gasPrice'] * 2

    @pytest.mark.parametrize(
        'raw_transaction, expected_hash',
        [
            (
                # address 0x39EEed73fb1D3855E90Cbd42f348b3D7b340aAA6
                '0xf8648085174876e8008252089439eeed73fb1d3855e90cbd42f348b3d7b340aaa601801ba0ec1295f00936acd0c2cb90ab2cdaacb8bf5e11b3d9957833595aca9ceedb7aada05dfc8937baec0e26029057abd3a1ef8c505dca2cdc07ffacb046d090d2bea06a',  # noqa: E501
                '0x1f80f8ab5f12a45be218f76404bda64d37270a6f4f86ededd0eb599f80548c13',
            ),
            (
                # private key 0x3c2ab4e8f17a7dea191b8c991522660126d681039509dc3bb31af7c9bdb63518
                # This is an unfunded account, but the transaction has a 0 gas price, so is valid.
                # It never needs to be mined, we just want the transaction hash back to confirm.
                HexBytes('0xf85f808082c35094d898d5e829717c72e7438bad593076686d7d164a80801ba005c2e99ecee98a12fbf28ab9577423f42e9e88f2291b3acc8228de743884c874a077d6bc77a47ad41ec85c96aac2ad27f05a039c4787fca8a1e5ee2d8c7ec1bb6a'),  # noqa: E501
                '0x98eeadb99454427f6aad7b558bac13e9d225512a6f5e5c11cf48e8d4067e51b5',
            ),
        ]
    )
    def test_eth_sendRawTransaction(self,
                                    web3,
                                    raw_transaction,
                                    funded_account_for_raw_txn,
                                    expected_hash):
        txn_hash = web3.platone.sendRawTransaction(raw_transaction)
        assert txn_hash == web3.toBytes(hexstr=expected_hash)

    def test_eth_call(self, web3, math_contract):
        coinbase = web3.platone.coinbase
        txn_params = math_contract._prepare_transaction(
            fn_name='add',
            fn_args=(7, 11),
            transaction={'from': coinbase, 'to': math_contract.address},
        )
        call_result = web3.platone.call(txn_params)
        assert is_string(call_result)
        result = decode_single('uint256', call_result)
        assert result == 18

    def test_eth_call_with_0_result(self, web3, math_contract):
        coinbase = web3.platone.coinbase
        txn_params = math_contract._prepare_transaction(
            fn_name='add',
            fn_args=(0, 0),
            transaction={'from': coinbase, 'to': math_contract.address},
        )
        call_result = web3.platone.call(txn_params)
        assert is_string(call_result)
        result = decode_single('uint256', call_result)
        assert result == 0

    def test_eth_estimateGas(self, web3, unlocked_account_dual_type):
        gas_estimate = web3.platone.estimateGas({
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
        })
        assert is_integer(gas_estimate)
        assert gas_estimate > 0

    def test_eth_getBlockByHash(self, web3, empty_block):
        block = web3.platone.getBlock(empty_block['hash'])
        assert block['hash'] == empty_block['hash']

    def test_eth_getBlockByHash_not_found(self, web3, empty_block):
        block = web3.platone.getBlock(UNKNOWN_HASH)
        assert block is None

    def test_eth_getBlockByNumber_with_integer(self, web3, empty_block):
        block = web3.platone.getBlock(empty_block['number'])
        assert block['number'] == empty_block['number']

    def test_eth_getBlockByNumber_latest(self, web3, empty_block):
        current_block_number = web3.platone.blockNumber
        block = web3.platone.getBlock('latest')
        assert block['number'] == current_block_number

    def test_eth_getBlockByNumber_not_found(self, web3, empty_block):
        block = web3.platone.getBlock(12345)
        assert block is None

    def test_eth_getBlockByNumber_pending(self, web3, empty_block):
        current_block_number = web3.platone.blockNumber
        block = web3.platone.getBlock('pending')
        assert block['number'] == current_block_number + 1

    def test_eth_getBlockByNumber_earliest(self, web3, empty_block):
        genesis_block = web3.platone.getBlock(0)
        block = web3.platone.getBlock('earliest')
        assert block['number'] == 0
        assert block['hash'] == genesis_block['hash']

    def test_eth_getBlockByNumber_full_transactions(self, web3, block_with_txn):
        block = web3.platone.getBlock(block_with_txn['number'], True)
        transaction = block['transactions'][0]
        assert transaction['hash'] == block_with_txn['transactions'][0]

    def test_eth_getTransactionByHash(self, web3, mined_txn_hash):
        transaction = web3.platone.getTransaction(mined_txn_hash)
        assert is_dict(transaction)
        assert transaction['hash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionByHash_contract_creation(self,
                                                        web3,
                                                        math_contract_deploy_txn_hash):
        transaction = web3.platone.getTransaction(math_contract_deploy_txn_hash)
        assert is_dict(transaction)
        assert transaction['to'] is None, "to field is %r" % transaction['to']

    def test_eth_getTransactionFromBlockHashAndIndex(self, web3, block_with_txn, mined_txn_hash):
        transaction = web3.platone.getTransactionFromBlock(block_with_txn['hash'], 0)
        assert is_dict(transaction)
        assert transaction['hash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionFromBlockNumberAndIndex(self, web3, block_with_txn, mined_txn_hash):
        transaction = web3.platone.getTransactionFromBlock(block_with_txn['number'], 0)
        assert is_dict(transaction)
        assert transaction['hash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionByBlockHashAndIndex(self, web3, block_with_txn, mined_txn_hash):
        transaction = web3.platone.getTransactionByBlock(block_with_txn['hash'], 0)
        assert is_dict(transaction)
        assert transaction['hash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionByBlockNumberAndIndex(self, web3, block_with_txn, mined_txn_hash):
        transaction = web3.platone.getTransactionByBlock(block_with_txn['number'], 0)
        assert is_dict(transaction)
        assert transaction['hash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionReceipt_mined(self, web3, block_with_txn, mined_txn_hash):
        receipt = web3.platone.getTransactionReceipt(mined_txn_hash)
        assert is_dict(receipt)
        assert receipt['blockNumber'] == block_with_txn['number']
        assert receipt['blockHash'] == block_with_txn['hash']
        assert receipt['transactionIndex'] == 0
        assert receipt['transactionHash'] == HexBytes(mined_txn_hash)

    def test_eth_getTransactionReceipt_unmined(self, web3, unlocked_account_dual_type):
        txn_hash = web3.platone.sendTransaction({
            'from': unlocked_account_dual_type,
            'to': unlocked_account_dual_type,
            'value': 1,
            'gas': 21000,
            'gasPrice': web3.platone.gasPrice,
        })
        receipt = web3.platone.getTransactionReceipt(txn_hash)
        assert receipt is None

    def test_eth_getTransactionReceipt_with_log_entry(self,
                                                      web3,
                                                      block_with_txn_with_log,
                                                      emitter_contract,
                                                      txn_hash_with_log):
        receipt = web3.platone.getTransactionReceipt(txn_hash_with_log)
        assert is_dict(receipt)
        assert receipt['blockNumber'] == block_with_txn_with_log['number']
        assert receipt['blockHash'] == block_with_txn_with_log['hash']
        assert receipt['transactionIndex'] == 0
        assert receipt['transactionHash'] == HexBytes(txn_hash_with_log)

        assert len(receipt['logs']) == 1
        log_entry = receipt['logs'][0]

        assert log_entry['blockNumber'] == block_with_txn_with_log['number']
        assert log_entry['blockHash'] == block_with_txn_with_log['hash']
        assert log_entry['logIndex'] == 0
        assert is_same_address(log_entry['address'], emitter_contract.address)
        assert log_entry['transactionIndex'] == 0
        assert log_entry['transactionHash'] == HexBytes(txn_hash_with_log)

    def test_eth_getUncleByBlockHashAndIndex(self, web3):
        # TODO: how do we make uncles....
        pass

    def test_eth_getUncleByBlockNumberAndIndex(self, web3):
        # TODO: how do we make uncles....
        pass

    def test_eth_getCompilers(self, web3):
        with pytest.warns(DeprecationWarning):
            web3.platone.getCompilers()

    def test_eth_compileSolidity(self, web3):
        # TODO: do we want to test this?
        pass

    def test_eth_compileLLL(self, web3):
        # TODO: do we want to test this?
        pass

    def test_eth_compileSerpent(self, web3):
        # TODO: do we want to test this?
        pass

    def test_eth_newFilter(self, web3):
        filter = web3.platone.filter({})

        changes = web3.platone.getFilterChanges(filter.filter_id)
        assert is_list_like(changes)
        assert not changes

        logs = web3.platone.getFilterLogs(filter.filter_id)
        assert is_list_like(logs)
        assert not logs

        result = web3.platone.uninstallFilter(filter.filter_id)
        assert result is True

    def test_eth_newBlockFilter(self, web3):
        filter = web3.platone.filter('latest')
        assert is_string(filter.filter_id)

        changes = web3.platone.getFilterChanges(filter.filter_id)
        assert is_list_like(changes)
        assert not changes

        # TODO: figure out why this fails in go-ethereum
        # logs = web3.platone.getFilterLogs(filter.filter_id)
        # assert is_list_like(logs)
        # assert not logs

        result = web3.platone.uninstallFilter(filter.filter_id)
        assert result is True

    def test_eth_newPendingTransactionFilter(self, web3):
        filter = web3.platone.filter('pending')
        assert is_string(filter.filter_id)

        changes = web3.platone.getFilterChanges(filter.filter_id)
        assert is_list_like(changes)
        assert not changes

        # TODO: figure out why this fails in go-ethereum
        # logs = web3.platone.getFilterLogs(filter.filter_id)
        # assert is_list_like(logs)
        # assert not logs

        result = web3.platone.uninstallFilter(filter.filter_id)
        assert result is True

    def test_eth_getLogs_without_logs(self, web3, block_with_txn_with_log):
        # Test with block range

        filter_params = {
            "fromBlock": 0,
            "toBlock": block_with_txn_with_log['number'] - 1,
        }
        result = web3.platone.getLogs(filter_params)
        assert len(result) == 0

        # the range is wrong
        filter_params = {
            "fromBlock": block_with_txn_with_log['number'],
            "toBlock": block_with_txn_with_log['number'] - 1,
        }
        result = web3.platone.getLogs(filter_params)
        assert len(result) == 0

        # Test with `address`

        # filter with other address
        filter_params = {
            "fromBlock": 0,
            "address": UNKNOWN_ADDRESS,
        }
        result = web3.platone.getLogs(filter_params)
        assert len(result) == 0

        # Test with multiple `address`

        # filter with other address
        filter_params = {
            "fromBlock": 0,
            "address": [UNKNOWN_ADDRESS, UNKNOWN_ADDRESS],
        }
        result = web3.platone.getLogs(filter_params)
        assert len(result) == 0

    def test_eth_getLogs_with_logs(
            self,
            web3,
            block_with_txn_with_log,
            emitter_contract_address,
            txn_hash_with_log):

        def assert_contains_log(result):
            assert len(result) == 1
            log_entry = result[0]
            assert log_entry['blockNumber'] == block_with_txn_with_log['number']
            assert log_entry['blockHash'] == block_with_txn_with_log['hash']
            assert log_entry['logIndex'] == 0
            assert is_same_address(log_entry['address'], emitter_contract_address)
            assert log_entry['transactionIndex'] == 0
            assert log_entry['transactionHash'] == HexBytes(txn_hash_with_log)

        # Test with block range

        # the range includes the block where the log resides in
        filter_params = {
            "fromBlock": block_with_txn_with_log['number'],
            "toBlock": block_with_txn_with_log['number'],
        }
        result = web3.platone.getLogs(filter_params)
        assert_contains_log(result)

        # specify only `from_block`. by default `to_block` should be 'latest'
        filter_params = {
            "fromBlock": 0,
        }
        result = web3.platone.getLogs(filter_params)
        assert_contains_log(result)

        # Test with `address`

        # filter with emitter_contract.address
        filter_params = {
            "fromBlock": 0,
            "address": emitter_contract_address,
        }

    def test_eth_getLogs_with_logs_topic_args(
            self,
            web3,
            block_with_txn_with_log,
            emitter_contract_address,
            txn_hash_with_log):
        def assert_contains_log(result):
            assert len(result) == 1
            log_entry = result[0]
            assert log_entry['blockNumber'] == block_with_txn_with_log['number']
            assert log_entry['blockHash'] == block_with_txn_with_log['hash']
            assert log_entry['logIndex'] == 0
            assert is_same_address(log_entry['address'], emitter_contract_address)
            assert log_entry['transactionIndex'] == 0
            assert log_entry['transactionHash'] == HexBytes(txn_hash_with_log)

        # Test with None event sig

        filter_params = {
            "fromBlock": 0,
            "topics": [
                None,
                '0x000000000000000000000000000000000000000000000000000000000000d431'],
        }

        result = web3.platone.getLogs(filter_params)
        assert_contains_log(result)

        # Test with None indexed arg
        filter_params = {
            "fromBlock": 0,
            "topics": [
                '0x057bc32826fbe161da1c110afcdcae7c109a8b69149f727fc37a603c60ef94ca',
                None],
        }
        result = web3.platone.getLogs(filter_params)
        assert_contains_log(result)

    def test_eth_getLogs_with_logs_none_topic_args(
            self,
            web3):
        # Test with None overflowing
        filter_params = {
            "fromBlock": 0,
            "topics": [None, None, None],
        }

        result = web3.platone.getLogs(filter_params)
        assert len(result) == 0

    def test_eth_call_old_contract_state(self, web3, math_contract, unlocked_account):
        start_block = web3.platone.getBlock('latest')
        block_num = start_block.number
        block_hash = start_block.hash

        math_contract.functions.increment().transact({'from': unlocked_account})

        # This isn't an incredibly convincing test since we can't mine, and
        # the default resolved block is latest, So if block_identifier was ignored
        # we would get the same result. For now, we mostly depend on core tests.
        # Ideas to improve this test:
        #  - Enable on-demand mining in more clients
        #  - Increment the math contract in all of the fixtures, and check the value in an old block
        block_hash_call_result = math_contract.functions.counter().call(block_identifier=block_hash)
        block_num_call_result = math_contract.functions.counter().call(block_identifier=block_num)
        latest_call_result = math_contract.functions.counter().call(block_identifier='latest')
        default_call_result = math_contract.functions.counter().call()
        pending_call_result = math_contract.functions.counter().call(block_identifier='pending')

        assert block_hash_call_result == 0
        assert block_num_call_result == 0
        assert latest_call_result == 0
        assert default_call_result == 0

        if pending_call_result != 1:
            raise AssertionError("pending call result was %d instead of 1" % pending_call_result)

    def test_eth_uninstallFilter(self, web3):
        filter = web3.platone.filter({})
        assert is_string(filter.filter_id)

        success = web3.platone.uninstallFilter(filter.filter_id)
        assert success is True

        failure = web3.platone.uninstallFilter(filter.filter_id)
        assert failure is False



ERROR_INFO = {
    # successful error code
    0: 'sendRawTransaction successful',
    # System error
    1: 'System error',
    # Object not found
    2: 'Object not found',
    # Invalid parameter
    3: 'Invalid parameter',

    # Node manager contract error code definition
    301001: 'Node manager contract internal error',
    301002: 'the node name is empty',
    301003: 'the length of node name is too long',
    301004: 'the length of node name is too short',
    301005: 'The node name has already used',
    301006: 'the value of node type is wrong',
    301007: 'the node owner is empty',
    301008: 'the length of node desc is too long',
    301009: 'the node publicKey is empty',
    301010: 'the node publicKey value format error',
    301011: 'The node publicKey has already used',
    301012: 'the node blsPubKey is empty',
    301013: 'the node blsPubKey value format error',
    301014: 'the node host address is empty',
    301015: 'The host address and port has already used',
    301016: 'the node rpc port is empty',
    301017: 'the node p2p port is empty',
    301018: 'the value of the audit status is wrong',
    301019: 'The node does not exists',
    301020: 'The node is not valid',
    301021: 'The node status is not unaudited',
    301022: 'The node status is not disabled',
    301023: 'The root node does not support update',
    301024: 'The node status value error',
    301025: 'The node owner does not exists',
    301026: 'The node owner is not valid',
    301027: 'The node owner has no permission',
    301028: 'The consensus node cannot be deleted',
    301029: 'The consensus node cannot be disabled',
    301030: 'get consensus node numbers error',
    301031: 'consensus node more than license definitioned number',

    # User manager contract error code definition'
    302001: 'User manager contract internal error',
    302002: 'the user address is empty',
    302003: 'the user name is empty',
    302004: 'the length of user name is too long',
    302005: 'the length of user name is too short',
    302006: 'the user mobile is empty',
    302007: 'the user mobile format is wrong',
    302008: 'the user email is empty',
    302009: 'the length of user email is too long',
    302010: 'the user email format is wrong',
    302011: 'the length of user desc is too long',
    302012: 'the value of the roles is wrong',
    302013: 'The user address has already registered',
    302014: 'The user name has already registered',
    302015: 'the value of the audit status is wrong',
    302016: 'The user does not exist',
    302017: 'The user is not a valid user',
    302018: 'The user status is not unaudited',
    302019: 'The user status is not disabled',
    302020: 'The caller does not exist',
    302021: 'The caller is not a valid user',
    302022: 'The caller has no permission',
    302023: 'The user has no roles',
    302024: 'The user status value error',

    # System parameter contract error code definition'
    303001: 'parameter error',
    303002: 'Internal error',
    303003: 'permission deny',

    # CNS contract error code definition'
    304001: 'Internal error',
    304002: 'Wrong contract name',
    304003: 'Wrong contract version',
    304004: 'Wrong contract address',
    304005: 'The caller is not the contract owner',
    304006: 'Contract name and version already registered',
    304007: 'Contract name already registered',
    304008: 'Compare contract version error',
    304009: 'Version is lower than registered version',
    304010: 'The contract is does not exists',

    # Ledger manager contract error code definition'
    305001: 'Wrong LedgerInfo Iuputs',
    305002: 'Internal error',
    305003: 'permission deny',
    305004: 'The ledgerName has already existed',
    305005: 'The ledgerName has already existed',
    305006: 'The ledgerName is empty',
    305007: 'The ledgerNode publicKey value format error',
    305008: 'The ledgerNode publicKey is empty',
    305009: 'The ledgerNode is does not exist',
    305010: 'The ledgerNodes do not have consensus node',
    305011: 'The ledgerNode BlsPublicKey value format error',
    305012: 'The ledgerNode BlsPublicKey is empty',
    305013: 'The JoinNode has already existed',
    305014: 'The quitNode is does not exist in ledger',
    305015: 'The ledgerName is too long',
    305016: 'Cannot quit the last node',
    305017: 'The ledger has deleted',
    305018: 'The ledgerNode is not valid',

    # Ledger node manager contract error code definition'
    306001: 'Internal error',
    306002: 'Empty public key',
    306003: 'Invalid public key',
    306004: 'Empty bls public key',
    306005: 'Invalid bls public key',
    306006: 'Permission deny',
    306007: 'Can not remove consensus node',
    306008: 'Node not found',
    306009: 'Invalid node type',
    306010: 'Empty address',
    306011: 'Cannot update the last consensus node type',
    306012: 'Node has already exist',

    307001: 'Invalid cipher data',
    307002: 'Re-encrypt cipher failed',
}
